# Copyright contributors to the Terratorch project

from __future__ import annotations

import asyncio
import logging
import os
from pathlib import Path
import copy

from collections.abc import Sequence
from typing import Any, Optional, Union

import rasterio
import torch
from vllm.config import VllmConfig
from vllm.entrypoints.pooling.pooling.protocol import IOProcessorRequest, IOProcessorResponse
from vllm.inputs.data import PromptType
from vllm.outputs import PoolingRequestOutput
from vllm.plugins.io_processors.interface import IOProcessor, IOProcessorInput, IOProcessorOutput

from terratorch.tasks.tiled_inference import generate_tiled_inference_output, prepare_tiled_inference_input
from terratorch.vllm.plugins import generate_datamodule
from terratorch.cli_tools import write_tiff
from terratorch.vllm.utils import check_vllm_version
from .utils import download_file_async, get_filename_from_url, path_or_tmpdir, to_base64_tiff

from .types import PluginConfig, RequestData, RequestOutput, TerramindSegmentationRequestInfo, TiledInferenceParameters

logger = logging.getLogger(__name__)


class TerramindSegmentationIOProcessor(IOProcessor):
    """vLLM IOProcessor for Terramind segmentation tasks

    This class instantiates an IO Processor plugin for vLLM for pre/post processing of multimodal data
    to be used with Terramind in Segmentation tasks.
    This plugin accepts multimodal data in the format of a url, or a directory path.
    Similarly, it can generate GeoTiff images is the form of a base64 encoded string or a file path.

    The plugin accepts and returns data in various formats and can be configured via the below environment variable:
        TERRATORCH_SEGMENTATION_IO_PROCESSOR_CONFIG
    This variable is to be set while starting the vLLM instance.
    The plugins configurable variables are:
    - output_path (String): Default path for storing output files when requesting output in 'path' mode. It is is ignored otherwise.
    The full schema of the plugin configuration can be found in vllm.plugins.segmentation.types.PluginConfig

    Once instantiated from the vLLM side, the plugin is automatically used when performing inference requests to the
    '/pooling' endpoint of a vLLM instance.
    """

    def __init__(self, vllm_config: VllmConfig):

        super().__init__(vllm_config)

        self.model_config = vllm_config.model_config.hf_config.to_dict()["pretrained_cfg"]

        if not "data" in self.model_config:
            raise ValueError("The model config does not contain the Terratorch datamodule configuration")

        plugin_config_string = os.getenv("TERRATORCH_SEGMENTATION_IO_PROCESSOR_CONFIG", "{}")

        self.plugin_config = PluginConfig.model_validate_json(plugin_config_string)

        self.tiled_inference_parameters = self._init_tiled_inference_parameters_info()
        self.batch_size = 1
        self.requests_cache: dict[str, TerramindSegmentationRequestInfo] = {}

    def _init_tiled_inference_parameters_info(self) -> TiledInferenceParameters:
        if "tiled_inference_parameters" in self.model_config["model"]["init_args"]:
            tiled_inf_param_dict = self.model_config["model"]["init_args"]["tiled_inference_parameters"]
            if not all(["h_crop" in tiled_inf_param_dict, "w_crop" in tiled_inf_param_dict]):
                if "crop" in tiled_inf_param_dict:
                    tiled_inf_param_dict["h_crop"] = tiled_inf_param_dict["crop"]
                    tiled_inf_param_dict["w_crop"] = tiled_inf_param_dict["crop"]
                    del tiled_inf_param_dict["crop"]
                else:
                    raise ValueError(
                        f"Expect 'crop' (or 'h_crop' and 'w_crop') in tiled_inference_parameters "
                        f"but got {tiled_inf_param_dict}"
                    )
            if "stride" in tiled_inf_param_dict:
                tiled_inf_param_dict["h_stride"] = tiled_inf_param_dict["stride"]
                tiled_inf_param_dict["w_stride"] = tiled_inf_param_dict["stride"]
                del tiled_inf_param_dict["stride"]
        else:
            tiled_inf_param_dict = {}

        return TiledInferenceParameters(**tiled_inf_param_dict)

    def _get_datamodule_config(self) -> dict:
        data_module_config = copy.deepcopy(self.model_config["data"])

        if "ImpactMeshDataModule" in data_module_config["class_path"]:
            # This is so that we can put the input data in a folder with an arbitrary name.
            # However, this requires for the means and stds to be included in the model configuration
            data_module_config["init_args"]["label_grep"] = ""

        return data_module_config

    async def _download_input_data(self, dataset_path: str, request_data: dict):

        # I am assuming the user to provide me with one url for each input modality
        download_tasks = []
        # with TemporaryDirectory(delete=False) as temp_dir:
        dir_path = Path(dataset_path)
        for modality, url in request_data.items():
            modality_dir = dir_path / modality
            modality_dir.mkdir()
            dest_file = modality_dir / get_filename_from_url(url)
            task = asyncio.create_task(download_file_async(url, dest_file))
            download_tasks.append(task)

        await asyncio.gather(*download_tasks)

    def parse_request(self, request: Any) -> IOProcessorInput:
        if type(request) is dict:
            image_prompt = RequestData(**request)
            return image_prompt
        if isinstance(request, IOProcessorRequest):
            if not hasattr(request, "data"):
                raise ValueError("missing 'data' field in OpenAIBaseModel Request")

            request_data = request.data

            if type(request_data) is dict:
                return RequestData(**request_data)
            else:
                raise ValueError("Unable to parse the request data")

        raise ValueError("Unable to parse request")

    def output_to_response(self, plugin_output: IOProcessorOutput) -> IOProcessorResponse:
        return IOProcessorResponse(
            request_id=plugin_output.request_id,
            data=plugin_output,
        )

    def pre_process(
        self,
        prompt: IOProcessorInput,
        request_id: Optional[str] = None,
        **kwargs,
    ) -> Union[PromptType, Sequence[PromptType]]:
        # Just run the async function froma. synchronous context.
        # Since we are already in the vLLM server event loop we use that one.
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.pre_process_async(prompt, request_id, **kwargs))

    async def pre_process_async(
        self,
        prompt: IOProcessorInput,
        request_id: Optional[str] = None,
        **kwargs,
    ) -> Union[PromptType, Sequence[PromptType]]:

        request_data = RequestData.model_validate(prompt)

        input_data_format = request_data.data_format

        # Validate out_path if provided and out_data_format is "path"
        if request_data.out_data_format == "path" and request_data.out_path:
            if not os.path.exists(request_data.out_path):
                raise ValueError(f"The output path '{request_data.out_path}' does not exist")
            if not os.access(request_data.out_path, os.W_OK):
                raise ValueError(f"The output path '{request_data.out_path}' is not writable")

        datamodule_config = self._get_datamodule_config()

        with path_or_tmpdir(request_data) as dataset_path:
            if input_data_format == "url":
                await self._download_input_data(dataset_path, request_data=request_data.data)

            # set the datamodule data_root to where the dataset is located
            datamodule_config["init_args"]["predict_data_root"] = dataset_path
            datamodule = generate_datamodule(datamodule_config)
            datamodule.batch_size = 1
            datamodule.setup("predict")

            # process the input files into Tensors
            data_loader = datamodule.predict_dataloader()
            data = list(data_loader)[0]

            # We can only retrieve metadata if the dataset contains a tiff file that we can load
            if "filename" in data:
                input_image_path = data["filename"][0]
                with rasterio.open(input_image_path, "r") as src:
                    metadata = src.meta
            else:
                logger.warning("No metadata can be loaded for the provided input")
                metadata = {}

        # Split the input in tiles depending on the tiled inference parameters
        input_data = datamodule.aug(data)["image"]
        prompt_data, tensor_reshape_fn, input_batch_size, h_img, w_img, _, delta = prepare_tiled_inference_input(
            input_data,
            **self.tiled_inference_parameters.model_dump(exclude={"average_patches"}),
        )

        prompts = []
        for tile in prompt_data:
            reshaped_tile = tensor_reshape_fn(tile.input_data)
            # TODO: Check if there's a better way of getting the data in the correct data type ouf of the box.
            multi_modal_data = {mod: tensor.to(torch.float16) for mod, tensor in reshaped_tile.items()}

            # after v0.14.0 vLLM has changed the input structure for multimodal data
            if check_vllm_version("0.14.0", ">"):
                multi_modal_data = {"image": multi_modal_data}

            prompt = {"prompt_token_ids": [1], "multi_modal_data": multi_modal_data}

            prompts.append(prompt)

        # if no request_id is passed this means that the plugin is used with vlLM
        # in offline sync mode. Therefore, we assume that one request at a time is being processed
        if not request_id:
            request_id = "offline"
        self.requests_cache[request_id] = TerramindSegmentationRequestInfo(
            out_data_format=request_data.out_data_format,
            out_path=request_data.out_path,
            dataset_path=dataset_path,
            prompt_data=prompt_data,
            h_img=h_img,
            w_img=w_img,
            input_batch_size=input_batch_size,
            metadata=metadata,
            filename=data["filename"][0],
            delta=delta,
        )

        return prompts

    def post_process(
        self,
        model_output: Sequence[PoolingRequestOutput],
        request_id: Optional[str] = None,
        **kwargs,
    ) -> IOProcessorOutput:

        if not request_id:
            request_id = "offline"

        if request_id and (request_id in self.requests_cache):
            request_info = self.requests_cache[request_id]
            del self.requests_cache[request_id]

        output_format = request_info.out_data_format

        model_outputs = [output.outputs.data.squeeze(0) for output in model_output]
        outputs = list(zip(request_info.prompt_data, model_outputs, strict=True))
        output = generate_tiled_inference_output(
            outputs=outputs,
            input_batch_size=request_info.input_batch_size,
            h_img=request_info.h_img,
            w_img=request_info.w_img,
            delta=request_info.delta,
            padding=self.tiled_inference_parameters.padding,
            average_patches=self.tiled_inference_parameters.average_patches,
        )

        prediction = output.squeeze(0).argmax(dim=0).numpy()

        metadata = request_info.metadata

        ret: str
        if output_format == "path":
            # Use out_path from request if provided, otherwise use plugin config output_path
            output_dir = request_info.out_path if request_info.out_path else self.plugin_config.output_path
            out_file_path = Path(output_dir) / f"{Path(request_info.filename).stem}_prediction.tif"
            write_tiff(prediction, out_file_path, metadata)
            ret = str(out_file_path.resolve())
        elif output_format == "b64_json":
            ret = to_base64_tiff(prediction, metadata=metadata)

        return RequestOutput(
            data_format=request_info.out_data_format,
            data=ret,
            request_id=request_id,
        )



# Copyright contributors to the Terratorch project

import warnings
from terratorch.models.heads.scalar_head import ScalarHead
from terratorch.models.heads.regression_head import RegressionHead
from terratorch.models.heads.segmentation_head import SegmentationHead

# TODO: Remove in a version v1.3 or later
class ClassificationHead(ScalarHead):
    def __init__(self, *args, **kwargs):
        warnings.warn("ClassificationHead is deprecated. Use ScalarHead instead", DeprecationWarning)
        super().__init__(*args, **kwargs)

__all__ = ["ScalarHead", "RegressionHead", "SegmentationHead"]



# Copyright contributors to the Terratorch project

"""Ugly proxy objects so parsing config file works with transforms.

    These are necessary since, for LightningCLI to instantiate arguments as
    objects from the config, they must have type annotations

    In TorchGeo, `transforms` is passed in **kwargs, so it has no type annotations!
    To get around that, we create these wrappers that have transforms type annotated.
    They create the transforms and forward all method and attribute calls to the
    original TorchGeo datamodule.

    Additionally, TorchGeo datasets pass the data to the transforms callable
    as a dict, and as a tensor.

    Albumentations expects this data not as a dict but as different key-value
    arguments, and as numpy. We handle that conversion here. 
"""
from collections.abc import Callable
from typing import Any

import numpy as np
from albumentations import BasicTransform
from torch import Tensor
from torchgeo.datamodules import GeoDataModule, NonGeoDataModule

from terratorch.datasets.transforms import albumentations_to_callable_with_dict

ALBUMENTATIONS_TARGETS = ["image", "mask"]


def build_callable_transform_from_torch_tensor(
    callable_transform: Callable[[dict[str, Tensor]], dict[str, Tensor]],
) -> Callable[[dict[str, np.ndarray]], dict[str, np.ndarray]]:
    # Take a function from dicts of str->Tensor to dicts of str->Tensor
    # Return a function from dicts of str->np to dicts of str->np
    # Additionally permute the image back to channels last

    def transforms_from_torch_tensor(tensor_dict: dict[str, Tensor]):
        numpy_dict = {k: (v.numpy() if k in ALBUMENTATIONS_TARGETS else v) for k, v in tensor_dict.items()}
        numpy_dict["image"] = np.moveaxis(numpy_dict["image"], 0, -1)  # image to channels last
        return callable_transform(numpy_dict)

    return transforms_from_torch_tensor


class TorchNonGeoDataModule(NonGeoDataModule):
    """Proxy object for using NonGeo data modules defined by TorchGeo.

    Allows for transforms to be defined and passed using config files.
    The only reason this class exists is so that we can annotate the transforms argument with a type.
    This is required for lightningcli and config files.
    As such, all getattr and setattr will be redirected to the underlying class.
    """

    def __init__(
        self,
        cls: type[NonGeoDataModule],
        batch_size: int | None = None,
        num_workers: int = 0,
        transforms: None | list[Any] = None,
        **kwargs: Any,
    ):
        """Constructor

        Args:
            cls (type[NonGeoDataModule]): TorchGeo DataModule class to be instantiated
            batch_size (int | None, optional): batch_size. Defaults to None.
            num_workers (int, optional): num_workers. Defaults to 0.
            transforms (None | list[BasicTransform], optional): List of Albumentations Transforms.
                Should enc with ToTensorV2. Defaults to None.
            **kwargs (Any): Arguments passed to instantiate `cls`.
        """
        if batch_size is not None:
            kwargs["batch_size"] = batch_size
        if transforms is not None:
            transforms_as_callable = albumentations_to_callable_with_dict(transforms)
            kwargs["transforms"] = build_callable_transform_from_torch_tensor(transforms_as_callable)
        # self.__dict__["datamodule"] = cls(num_workers=num_workers, **kwargs)
        self._proxy = cls(num_workers=num_workers, **kwargs)
        super().__init__(self._proxy.dataset_class)  # dummy arg

    @property
    def collate_fn(self):
        return self._proxy.collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._proxy.collate_fn = value

    def setup(self, stage: str):
        return self._proxy.setup(stage)

    def train_dataloader(self):
        return self._proxy.train_dataloader()

    def val_dataloader(self):
        return self._proxy.val_dataloader()

    def test_dataloader(self):
        return self._proxy.test_dataloader()

    def predict_dataloader(self):
        return self._proxy.predict_dataloader()


class TorchGeoDataModule(GeoDataModule):
    """Proxy object for using Geo data modules defined by TorchGeo.

    Allows for transforms to be defined and passed using config files.
    The only reason this class exists is so that we can annotate the transforms argument with a type.
    This is required for lightningcli and config files.
    As such, all getattr and setattr will be redirected to the underlying class.
    """

    def __init__(
        self,
        cls: type[GeoDataModule],
        batch_size: int | None = None,
        num_workers: int = 0,
        transforms: None | list[Any] = None,
        **kwargs: Any,
    ):
        """Constructor

        Args:
            cls (type[GeoDataModule]): TorchGeo DataModule class to be instantiated
            batch_size (int | None, optional): batch_size. Defaults to None.
            num_workers (int, optional): num_workers. Defaults to 0.
            transforms (None | list[BasicTransform], optional): List of Albumentations Transforms.
                Should enc with ToTensorV2. Defaults to None.
            **kwargs (Any): Arguments passed to instantiate `cls`.
        """
        if batch_size is not None:
            kwargs["batch_size"] = batch_size
        if transforms is not None:
            transforms_as_callable = albumentations_to_callable_with_dict(transforms)
            kwargs["transforms"] = build_callable_transform_from_torch_tensor(transforms_as_callable)
        # self.__dict__["datamodule"] = cls(num_workers=num_workers, **kwargs)
        self._proxy = cls(num_workers=num_workers, **kwargs)
        super().__init__(self._proxy.dataset_class)  # dummy arg

    @property
    def collate_fn(self):
        return self._proxy.collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._proxy.collate_fn = value

    @property
    def patch_size(self):
        return self._proxy.patch_size

    @property
    def length(self):
        return self._proxy.length

    def setup(self, stage: str):
        return self._proxy.setup(stage)

    def train_dataloader(self):
        return self._proxy.train_dataloader()

    def val_dataloader(self):
        return self._proxy.val_dataloader()

    def test_dataloader(self):
        return self._proxy.test_dataloader()

    def predict_dataloader(self):
        return self._proxy.predict_dataloader()

    def transfer_batch_to_device(self, batch, device, dataloader_idx):
        return self._proxy.predict_dataloader(batch, device, dataloader_idx)



from collections.abc import Sequence
from typing import Any

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from kornia.augmentation import AugmentationSequential

from terratorch.datamodules.generic_multimodal_data_module import wrap_in_compose_is_list
from terratorch.datasets import Landslide4SenseNonGeo

MEANS = {
    "COASTAL AEROSOL": -0.4914,
    "BLUE": -0.3074,
    "GREEN": -0.1277,
    "RED": -0.0625,
    "RED_EDGE_1": 0.0439,
    "RED_EDGE_2": 0.0803,
    "RED_EDGE_3": 0.0644,
    "NIR_BROAD": 0.0802,
    "WATER_VAPOR": 0.3000,
    "CIRRUS": 0.4082,
    "SWIR_1": 0.0823,
    "SWIR_2": 0.0516,
    "SLOPE": 0.3338,
    "DEM": 0.7819,
}

STDS = {
    "COASTAL AEROSOL": 0.9325,
    "BLUE": 0.8775,
    "GREEN": 0.8860,
    "RED": 0.8869,
    "RED_EDGE_1": 0.8857,
    "RED_EDGE_2": 0.8418,
    "RED_EDGE_3": 0.8354,
    "NIR_BROAD": 0.8491,
    "WATER_VAPOR": 0.9061,
    "CIRRUS": 1.6072,
    "SWIR_1": 0.8848,
    "SWIR_2": 0.9232,
    "SLOPE": 0.9018,
    "DEM": 1.2913,
}


class Landslide4SenseNonGeoDataModule(NonGeoDataModule):
    """NonGeo LightningDataModule implementation for Landslide4Sense dataset."""

    def __init__(
        self,
        data_root: str,
        batch_size: int = 4,
        num_workers: int = 0,
        bands: Sequence[str] = Landslide4SenseNonGeo.all_band_names,
        train_transform: A.Compose | None | list = None,
        val_transform: A.Compose | None | list = None,
        test_transform: A.Compose | None | list = None,
        predict_transform: A.Compose | None | list = None,
        aug: AugmentationSequential = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the Landslide4SenseNonGeoDataModule.

        Args:
            data_root (str): Root directory of the dataset.
            batch_size (int, optional): Batch size for data loaders. Defaults to 4.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            bands (Sequence[str], optional): List of band names to use. Defaults to Landslide4SenseNonGeo.all_band_names.
            train_transform (A.Compose | None | list, optional): Transformations for training data.
            val_transform (A.Compose | None | list, optional): Transformations for validation data.
            test_transform (A.Compose | None | list, optional): Transformations for testing data.
            predict_transform (A.Compose | None | list, optional): Transformations for prediction data.
            aug (AugmentationSequential, optional): Augmentation pipeline; if None, applies normalization using computed means and stds.
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(Landslide4SenseNonGeo, batch_size, num_workers, **kwargs)
        self.data_root = data_root

        self.means = [MEANS[b] for b in bands]
        self.stds = [STDS[b] for b in bands]
        self.bands = bands
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.predict_transform = wrap_in_compose_is_list(predict_transform)
        self.aug = (
            AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=None) if aug is None else aug
        )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either fit, validate, test, or predict.
        """
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                split="train",
                data_root=self.data_root,
                transform=self.train_transform,
                bands=self.bands
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val",
                data_root=self.data_root,
                transform=self.val_transform,
                bands=self.bands
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",
                data_root=self.data_root,
                transform=self.test_transform,
                bands=self.bands
            )
        if stage in ["predict"]:
            self.predict_dataset = self.dataset_class(
                split="test",
                data_root=self.data_root,
                transform=self.predict_transform,
                bands=self.bands
            )



from torchgeo.datasets import VHR10
import pandas as pd
import numpy as np

from torchgeo.datasets.utils import (
    Path,
    check_integrity,
    download_and_extract_archive,
    download_url,
    lazy_import,
    percentile_normalization,
)

from collections.abc import Callable
from typing import Any, ClassVar

from torch import Tensor
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib import patches
import pdb


class mVHR10(VHR10):
    def __init__(
        self,
        second_level_split="train",
        second_level_split_proportions = (0.90, 0.05, 0.05),
        boxes_output_tag='boxes',
        labels_output_tag='labels',
        masks_output_tag='masks',
        scores_output_tag='scores',
        *args,
        **kwargs) -> None:
        """Initialize a new m-VHR-10 dataset instance. On the original VHR10 splits (positive or negative) it creates train/val/test splits.

        Args:
            second_level_split: either "train", "val", "test"
            second_level_split_proportions: proportion for train, val and test as a tuple. It has to sum to 1.
        Raises:
            AssertionError: if ``split`` argument is invalid
            AssertionError: if ``second_level_split_proportions`` is not composed of 3 elements and sums up to 1.
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: if ``split="positive"`` and pycocotools is
                not installed.
        """
        super().__init__(*args, **kwargs)

        assert second_level_split in ['train', 'val', 'test']
        assert len(second_level_split_proportions) == 3
        assert np.sum(second_level_split_proportions) == 1
        self.second_level_split = second_level_split
        # pdb.set_trace()
        df = pd.DataFrame({"ids": self.ids})
        df = df.sort_values("ids")
        df = df.sample(frac=1, random_state=123)
        index_train_end = int(df.shape[0]*second_level_split_proportions[0])
        index_val_end = int(df.shape[0]*(second_level_split_proportions[0] + second_level_split_proportions[1]))

        self.boxes_output_tag = boxes_output_tag
        self.labels_output_tag = labels_output_tag
        self.masks_output_tag = masks_output_tag
        self.scores_output_tag = scores_output_tag

        if self.second_level_split == 'train':
            self.ids = list(df['ids'].values[:index_train_end])
        elif self.second_level_split == 'val':
            self.ids = list(df['ids'].values[index_train_end:index_val_end])
        else:
            self.ids = list(df['ids'].values[index_val_end:])

    def __getitem__(self, index: int) -> dict[str, Any]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        
        id_ = self.ids[index] + 1

        sample: dict[str, Any] = {
            'image': self._load_image(id_),
            'label': self._load_target(id_),
        }

        if sample['label']['annotations']:
            sample = self.coco_convert(sample)
            sample[self.boxes_output_tag] = sample['label']['boxes']
            sample[self.masks_output_tag] = sample['label']['masks']
            sample[self.labels_output_tag] = sample['label']['labels']
            if self.labels_output_tag != 'label':
                del sample['label']
        
        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample
    
    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        show_feats: str | None = 'both',
        box_alpha: float = 0.7,
        mask_alpha: float = 0.7,
        confidence_score = 0.5
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle
            show_titles: flag indicating whether to show titles above each panel
            show_feats: optional string to pick features to be shown: boxes, masks, both
            box_alpha: alpha value of box
            mask_alpha: alpha value of mask

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            AssertionError: if ``show_feats`` argument is invalid
            DependencyNotFoundError: If plotting masks and scikit-image is not installed.

        .. versionadded:: 0.4
        """
        assert show_feats in {'boxes', 'masks', 'both'}
        image = percentile_normalization(sample['image'].permute(1, 2, 0).numpy())

        if self.split == 'negative':
            fig, axs = plt.subplots(squeeze=False)
            axs[0, 0].imshow(image)
            axs[0, 0].axis('off')

            if suptitle is not None:
                plt.suptitle(suptitle)
            return fig

        if show_feats != 'boxes':
            skimage = lazy_import('skimage')

        boxes = sample[self.boxes_output_tag].cpu().numpy()
        labels = sample[self.labels_output_tag].cpu().numpy()

        if self.masks_output_tag in sample:
            masks = [mask.squeeze().cpu().numpy() for mask in sample[self.masks_output_tag]]

        n_gt = len(boxes)

        ncols = 1
        show_predictions = 'prediction_' + self.labels_output_tag in sample

        if show_predictions:
            show_pred_boxes = False
            show_pred_masks = False
            prediction_labels = sample['prediction_' + self.labels_output_tag].numpy()
            prediction_scores = sample['prediction_' + self.scores_output_tag].numpy()
            if 'prediction_' + self.boxes_output_tag in sample:
                prediction_boxes = sample['prediction_' + self.boxes_output_tag].numpy()
                show_pred_boxes = True
            if 'prediction_' + self.masks_output_tag in sample:
                prediction_masks = sample['prediction_' + self.masks_output_tag].numpy()
                show_pred_masks = True

            n_pred = len(prediction_labels)
            ncols += 1

        # Display image
        fig, axs = plt.subplots(ncols=ncols, squeeze=False, figsize=(ncols * 10, 13))
        axs[0, 0].imshow(image)
        axs[0, 0].axis('off')

        cm = plt.get_cmap('gist_rainbow')
        for i in range(n_gt):
            class_num = labels[i]
            color = cm(class_num / len(self.categories))

            # Add bounding boxes
            x1, y1, x2, y2 = boxes[i]
            if show_feats in {'boxes', 'both'}:
                r = patches.Rectangle(
                    (x1, y1),
                    x2 - x1,
                    y2 - y1,
                    linewidth=2,
                    alpha=box_alpha,
                    linestyle='dashed',
                    edgecolor=color,
                    facecolor='none',
                )
                axs[0, 0].add_patch(r)

            # Add labels
            label = self.categories[class_num]
            caption = label
            axs[0, 0].text(
                x1, y1 - 8, caption, color='white', size=11, backgroundcolor='none'
            )

            # Add masks
            if show_feats in {'masks', 'both'} and self.masks_output_tag in sample:
                mask = masks[i]
                contours = skimage.measure.find_contours(mask, 0.5)
                for verts in contours:
                    verts = np.fliplr(verts)
                    p = patches.Polygon(
                        verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                    )
                    axs[0, 0].add_patch(p)

            if show_titles:
                axs[0, 0].set_title('Ground Truth')

        if show_predictions:
            axs[0, 1].imshow(image)
            axs[0, 1].axis('off')
            for i in range(n_pred):
                score = prediction_scores[i]
                if score < confidence_score:
                    continue

                class_num = prediction_labels[i]
                color = cm(class_num / len(self.categories))

                if show_pred_boxes:
                    # Add bounding boxes
                    x1, y1, x2, y2 = prediction_boxes[i]
                    r = patches.Rectangle(
                        (x1, y1),
                        x2 - x1,
                        y2 - y1,
                        linewidth=2,
                        alpha=box_alpha,
                        linestyle='dashed',
                        edgecolor=color,
                        facecolor='none',
                    )
                    axs[0, 1].add_patch(r)

                    # Add labels
                    label = self.categories[class_num]
                    caption = f'{label} {score:.3f}'
                    axs[0, 1].text(
                        x1,
                        y1 - 8,
                        caption,
                        color='white',
                        size=11,
                        backgroundcolor='none',
                    )

                # Add masks
                if show_pred_masks:

                    mask = prediction_masks[i][0]
                    contours = skimage.measure.find_contours(mask, 0.5)
                    for verts in contours:
                        verts = np.fliplr(verts)
                        p = patches.Polygon(
                            verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                        )
                        axs[0, 1].add_patch(p)

            if show_titles:
                axs[0, 1].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        plt.tight_layout()

        return fig



# Copyright contributors to the Terratorch project


import logging

import terratorch.models.necks  # register necks  # noqa: F401
from terratorch.models.clay1_5_model_factory import Clay1_5ModelFactory
from terratorch.models.clay_model_factory import ClayModelFactory
from terratorch.models.encoder_decoder_factory import EncoderDecoderFactory
from terratorch.models.full_model_factory import FullModelFactory
from terratorch.models.generic_model_factory import GenericModelFactory
from terratorch.models.generic_unet_model_factory import GenericUnetModelFactory
from terratorch.models.object_detection_model_factory import ObjectDetectionModelFactory
from terratorch.models.prithvi_model_factory import PrithviModelFactory
from terratorch.models.satmae_model_factory import SatMAEModelFactory
from terratorch.models.smp_model_factory import SMPModelFactory
from terratorch.models.timm_model_factory import TimmModelFactory

try:
    granitewcx = True
    from terratorch.models.wxc_model_factory import WxCModelFactory
except ImportError:
    logging.getLogger("terratorch").debug("granitewxc not installed, please use pip install granitewxc")
    granitewcx = False

__all__ = (
    "Clay1_5ModelFactory",
    "ClayModelFactory",
    "EncoderDecoderFactory",
    "FullModelFactory",
    "GenericModelFactory",
    "GenericUnetModelFactory",
    "ObjectDetectionModelFactory",
    "PrithviModelFactory",
    "SMPModelFactory",
    "SatMAEModelFactory",
    "TimmModelFactory",
    "WxCModelFactory",
)

if granitewcx:
    __all__.__add__((WxCModelFactory,))



import gc
import os

import pytest
import torch

from terratorch.cli_tools import build_lightning_cli
from terratorch.registry import FULL_MODEL_REGISTRY
from terratorch.tasks import ReconstructionTask


@pytest.mark.parametrize("model_name", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_reconstruction_cli(model_name, case):
    command_list = [case, "-c", f"tests/resources/configs/manufactured-reconstruction_{model_name}.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()


@pytest.mark.parametrize("model_name", ["prithvi_eo_v1_100_mae"])
def test_prithvi_mae_reconstruction(model_name):
    model = FULL_MODEL_REGISTRY.build(
        model_name,
        pretrained=False,
    )

    input = torch.ones((1, 6, 224, 224))
    loss, reconstruction, mask = model(input)

    assert "loss" in loss
    assert reconstruction.shape == input.shape
    assert list(mask.shape) == [1, *reconstruction.shape[-2:]]

    gc.collect()


@pytest.mark.parametrize("model_name", ["terramind_v01_base_generate"])
def test_terramind_v01_generation(model_name):
    try:
        import diffusers
        import tokenizers
    except ImportError:
        pytest.skip("diffusers or tokenizers not installed")

    model = FULL_MODEL_REGISTRY.build(
        model_name,
        pretrained=False,
        modalities=["S2L2A", "LULC"],
        output_modalities=["S1GRD", "coords", "captions"],
        timesteps=1,
        standardize=True,
        offset={"S2L2A": 1},
    )

    # Test kwargs inputs
    output = model(S2L2A=torch.ones((1, 12, 224, 224)), LULC=torch.ones((1, 1, 224, 224)))

    model = FULL_MODEL_REGISTRY.build(
        model_name,
        pretrained=False,
        modalities=["S2L2A", "coords", "captions"],
        output_modalities=["S1GRD", "LULC", "captions"],
        timesteps=1,
    )

    input = {
        "S2L2A": torch.ones((1, 12, 224, 224)),
        "coords": torch.ones((1, 2)),
        "captions": ["This is a test"],
    }
    output = model(input)

    gc.collect()


@pytest.mark.parametrize("model_name", ["terramind_v1_base_generate"])
def test_terramind_generation(model_name):
    try:
        import diffusers
        import tokenizers
    except ImportError:
        pytest.skip("diffusers or tokenizers not installed")

    model = FULL_MODEL_REGISTRY.build(
        model_name,
        pretrained=False,
        modalities=["S2L2A", "coords"],
        output_modalities=["S1GRD", "LULC"],
        timesteps=1,
    )

    input = {
        "S2L2A": torch.ones((1, 12, 224, 224)),
        "coords": torch.ones((1, 2)),
    }
    output = model(input)

    gc.collect()



"""Comprehensive tests for PASTIS dataset to maximize coverage.

Tests cover:
- Dataset initialization with various parameters
- Semantic and instance segmentation targets
- Multiple satellite configurations (S2, S1A, S1D)
- Normalization on/off
- Class mapping
- Image truncation and padding
- Date handling and conversions
- Fold selection
- Transform application
- Plot functionality
- Error conditions
"""
from __future__ import annotations

import json
from datetime import datetime, timezone
from unittest.mock import Mock, patch

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import torch

from terratorch.datasets.pastis import PASTIS


@pytest.fixture
def dummy_pastis_data_full(tmp_path) -> str:
    """Create comprehensive dummy PASTIS dataset with all satellite types and instance annotations."""
    data_root = tmp_path / "pastis_full"
    data_root.mkdir()
    
    # Create metadata with multiple dates for S2, S1A, S1D
    df = pd.DataFrame({
        "ID_PATCH": [1, 2, 3, 4, 5],
        "Fold": [1, 1, 2, 3, 5],
        "dates-S2": [
            json.dumps({"0": "20180701", "1": "20180801", "2": "20180901"}),
            json.dumps({"0": "20180715"}),
            json.dumps({"0": "20180901", "1": "20181001"}),
            json.dumps({"0": "20180901"}),
            json.dumps({"0": "20180601", "1": "20180901", "2": "20181201"}),
        ],
        "dates-S1A": [
            json.dumps({"0": "20180705", "1": "20180805"}),
            json.dumps({"0": "20180720"}),
            json.dumps({"0": "20180905"}),
            json.dumps({"0": "20180910"}),
            json.dumps({"0": "20180610", "1": "20180910"}),
        ],
        "dates-S1D": [
            json.dumps({"0": "20180710"}),
            json.dumps({"0": "20180725"}),
            json.dumps({"0": "20180910"}),
            json.dumps({"0": "20180915"}),
            json.dumps({"0": "20180615"}),
        ],
    })
    gdf = gpd.GeoDataFrame(df, geometry=[None] * 5)
    meta_path = data_root / "metadata.geojson"
    gdf.to_file(meta_path, driver="GeoJSON")
    
    # Create normalization files for all satellites
    for satellite in ["S2", "S1A", "S1D"]:
        norm_dict = {}
        for f in range(1, 6):
            if satellite == "S2":
                norm_dict[f"Fold_{f}"] = {"mean": [0.3, 0.4, 0.5], "std": [0.1, 0.15, 0.2]}
            else:
                norm_dict[f"Fold_{f}"] = {"mean": [0.2, 0.3], "std": [0.05, 0.1]}
        norm_path = data_root / f"NORM_{satellite}_patch.json"
        with open(norm_path, "w") as file:
            json.dump(norm_dict, file)
    
    # Create data directories and files
    for satellite in ["S2", "S1A", "S1D"]:
        data_dir = data_root / f"DATA_{satellite}"
        data_dir.mkdir()
        for patch_id in [1, 2, 3, 4, 5]:
            if satellite == "S2":
                # S2: (time, channels, h, w) - 3 channels
                dummy_array = np.random.rand(3, 3, 128, 128).astype(np.float32)
            else:
                # S1: (time, channels, h, w) - 2 channels
                dummy_array = np.random.rand(2, 2, 128, 128).astype(np.float32)
            file_path = data_dir / f"{satellite}_{patch_id}.npy"
            np.save(file_path, dummy_array)
    
    # Create semantic annotations
    annotations_dir = data_root / "ANNOTATIONS"
    annotations_dir.mkdir()
    for patch_id in [1, 2, 3, 4, 5]:
        # Semantic target with class labels 0-17
        dummy_target = np.random.randint(0, 18, size=(1, 128, 128), dtype=np.int32)
        file_path = annotations_dir / f"TARGET_{patch_id}.npy"
        np.save(file_path, dummy_target)
    
    # Create instance annotations
    instance_dir = data_root / "INSTANCE_ANNOTATIONS"
    instance_dir.mkdir()
    for patch_id in [1, 2, 3, 4, 5]:
        # Heatmap (centerness)
        heatmap = np.random.rand(128, 128).astype(np.float32)
        np.save(instance_dir / f"HEATMAP_{patch_id}.npy", heatmap)
        
        # Instance IDs (parcels)
        instance_ids = np.zeros((128, 128), dtype=np.int32)
        instance_ids[:40, :40] = 1
        instance_ids[40:80, 40:80] = 2
        instance_ids[80:120, 80:120] = 3
        np.save(instance_dir / f"INSTANCES_{patch_id}.npy", instance_ids)
        
        # Zones (Voronoi partitioning)
        zones = np.zeros((128, 128), dtype=np.int32)
        zones[:64, :64] = 1
        zones[:64, 64:] = 2
        zones[64:, :64] = 3
        zones[64:, 64:] = 0
        np.save(instance_dir / f"ZONES_{patch_id}.npy", zones)
    
    return str(data_root)


class TestPASTISInitialization:
    """Test dataset initialization with various parameter combinations."""
    
    def test_basic_semantic_target(self, dummy_pastis_data_full):
        """Test basic initialization with semantic target."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            target="semantic",
            satellites=["S2"],
        )
        assert len(dataset) == 5
        assert dataset.target == "semantic"
        assert dataset.satellites == ["S2"]
        assert dataset.norm is not None
    
    def test_instance_target(self, dummy_pastis_data_full):
        """Test initialization with instance target."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=False,
            target="instance",
            satellites=["S2"],
        )
        assert dataset.target == "instance"
        assert dataset.norm is None
    
    def test_multiple_satellites(self, dummy_pastis_data_full):
        """Test initialization with multiple satellite configurations."""
        # S2 only
        dataset = PASTIS(data_root=dummy_pastis_data_full, satellites=["S2"])
        assert dataset.satellites == ["S2"]
        
        # S2 + S1A
        dataset = PASTIS(data_root=dummy_pastis_data_full, satellites=["S2", "S1A"])
        assert set(dataset.satellites) == {"S2", "S1A"}
        
        # All satellites
        dataset = PASTIS(data_root=dummy_pastis_data_full, satellites=["S2", "S1A", "S1D"])
        assert set(dataset.satellites) == {"S2", "S1A", "S1D"}
    
    def test_fold_selection(self, dummy_pastis_data_full):
        """Test selecting specific folds."""
        # Single fold
        dataset = PASTIS(data_root=dummy_pastis_data_full, folds=[1], satellites=["S2"])
        assert len(dataset) == 2  # Patches 1 and 2 are in fold 1
        
        # Multiple folds
        dataset = PASTIS(data_root=dummy_pastis_data_full, folds=[1, 2], satellites=["S2"])
        assert len(dataset) == 3  # Patches 1, 2, 3
        
        # All folds (None)
        dataset = PASTIS(data_root=dummy_pastis_data_full, folds=None, satellites=["S2"])
        assert len(dataset) == 5
    
    def test_class_mapping(self, dummy_pastis_data_full):
        """Test custom class mapping."""
        # Map all classes to binary (background vs crop)
        class_mapping = {i: 0 if i == 0 else 1 for i in range(18)}
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="semantic",
            class_mapping=class_mapping,
            satellites=["S2"],
        )
        assert dataset.class_mapping is not None
    
    def test_custom_reference_date(self, dummy_pastis_data_full):
        """Test custom reference date."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            reference_date="2018-01-01",
            satellites=["S2"],
        )
        expected_date = datetime(2018, 1, 1, tzinfo=timezone.utc)
        assert dataset.reference_date == expected_date
    
    def test_custom_date_interval(self, dummy_pastis_data_full):
        """Test custom date interval."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            date_interval=(-100, 300),
            satellites=["S2"],
        )
        assert len(dataset.date_range) == 400
        assert dataset.date_range[0] == -100
        assert dataset.date_range[-1] == 299
    
    def test_invalid_target_raises_error(self, dummy_pastis_data_full):
        """Test that invalid target raises ValueError."""
        with pytest.raises(ValueError, match="Target 'invalid' not recognized"):
            PASTIS(data_root=dummy_pastis_data_full, target="invalid", satellites=["S2"])
    
    def test_invalid_satellite_raises_error(self, dummy_pastis_data_full):
        """Test that invalid satellite raises ValueError."""
        with pytest.raises(ValueError, match="Satellite 'INVALID' not recognized"):
            PASTIS(data_root=dummy_pastis_data_full, satellites=["INVALID"])


class TestPASTISGetItem:
    """Test __getitem__ functionality with various configurations."""
    
    def test_semantic_target_getitem(self, dummy_pastis_data_full):
        """Test getting item with semantic target."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            target="semantic",
            satellites=["S2"],
        )
        sample = dataset[0]
        
        assert "image" in sample
        assert "mask" in sample
        assert "dates" in sample
        assert "S2" in sample
        
        # Check shapes
        assert sample["image"].shape[0] == 3  # timesteps
        assert sample["image"].shape[-1] == 3  # channels (transposed)
        assert sample["mask"].shape == (128, 128)
        assert "S2" in sample["dates"]
    
    def test_instance_target_getitem(self, dummy_pastis_data_full):
        """Test getting item with instance target."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=False,
            target="instance",
            satellites=["S2"],
        )
        sample = dataset[0]
        
        assert "mask" in sample
        # Instance target has 7 channels:
        # heatmap, instance_ids, pixel_to_object_mapping, size(h,w), object_semantic, pixel_semantic
        assert sample["mask"].shape[-1] == 7
        assert sample["mask"].dtype == np.float32
    
    def test_multiple_satellites_getitem(self, dummy_pastis_data_full):
        """Test getting item with multiple satellites."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2", "S1A", "S1D"],
        )
        sample = dataset[0]
        
        assert "S2" in sample
        assert "S1A" in sample
        assert "S1D" in sample
        assert "S2" in sample["dates"]
        assert "S1A" in sample["dates"]
        assert "S1D" in sample["dates"]
    
    def test_normalization_applied(self, dummy_pastis_data_full):
        """Test that normalization is applied correctly."""
        dataset_norm = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            satellites=["S2"],
        )
        dataset_no_norm = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=False,
            satellites=["S2"],
        )
        
        sample_norm = dataset_norm[0]
        sample_no_norm = dataset_no_norm[0]
        
        # Normalized data should have different values
        assert not np.allclose(sample_norm["S2"], sample_no_norm["S2"])
    
    def test_class_mapping_applied(self, dummy_pastis_data_full):
        """Test that class mapping is applied to semantic target."""
        class_mapping = {i: 0 if i < 5 else 1 for i in range(18)}
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="semantic",
            class_mapping=class_mapping,
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # All values should be 0 or 1 after mapping
        unique_vals = np.unique(sample["mask"])
        assert all(val in [0, 1] for val in unique_vals)
    
    def test_truncate_image(self, dummy_pastis_data_full):
        """Test image truncation."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            truncate_image=2,
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # Should truncate to last 2 timesteps
        assert sample["S2"].shape[0] == 2
        assert len(sample["dates"]["S2"]) == 2
    
    def test_pad_image(self, dummy_pastis_data_full):
        """Test image padding."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            pad_image=10,
            satellites=["S2"],
        )
        sample = dataset[1]  # Patch 2 has only 1 timestep
        
        # Should pad to 10 timesteps
        assert sample["S2"].shape[0] == 10
        assert len(sample["dates"]["S2"]) == 10
    
    def test_truncate_when_larger(self, dummy_pastis_data_full):
        """Test that truncation only happens when data is larger."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            truncate_image=5,  # Data has 3 timesteps, so no truncation
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # Should keep original 3 timesteps
        assert sample["S2"].shape[0] == 3
    
    def test_pad_when_smaller(self, dummy_pastis_data_full):
        """Test that padding only happens when data is smaller."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            pad_image=2,  # Data has 3 timesteps, so no padding
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # Should keep original 3 timesteps
        assert sample["S2"].shape[0] == 3
    
    def test_transform_applied(self, dummy_pastis_data_full):
        """Test that transform is applied to sample."""
        def custom_transform(**kwargs):
            # Modify image
            if "image" in kwargs:
                kwargs["image"] = kwargs["image"] * 2
            kwargs["custom_key"] = "custom_value"
            return kwargs
        
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            transform=custom_transform,
            satellites=["S2"],
        )
        sample = dataset[0]
        
        assert "custom_key" in sample
        assert sample["custom_key"] == "custom_value"
    
    def test_instance_annotations_with_zero_instance(self, dummy_pastis_data_full):
        """Test instance target processing with zero instance IDs."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="instance",
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # The instance processing should handle instance_id == 0 correctly
        mask = sample["mask"]
        assert mask.shape[-1] == 7
        
        # Check that size and semantic annotations are computed
        # For non-zero instances
        assert mask[..., 3:5].max() > 0  # Size should be non-zero for instances


class TestPASTISDateHandling:
    """Test date handling and conversion."""
    
    def test_get_dates(self, dummy_pastis_data_full):
        """Test get_dates method."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            reference_date="2018-09-01",
            satellites=["S2"],
        )
        
        # Get dates for patch 1 (ID_PATCH=1)
        dates = dataset.get_dates(1, "S2")
        
        # Should return dates relative to reference date
        assert isinstance(dates, np.ndarray)
        assert len(dates) > 0
    
    def test_date_conversion_from_string(self, dummy_pastis_data_full):
        """Test that string dates in metadata are converted correctly."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            reference_date="2018-09-01",
            satellites=["S2"],
        )
        
        sample = dataset[0]
        dates = sample["dates"]["S2"]
        
        # Dates should be torch tensors
        assert isinstance(dates, torch.Tensor)
        
        # Dates should be relative to reference date (in days)
        assert dates.dtype == torch.int64 or dates.dtype == torch.long


class TestPASTISPlotting:
    """Test plot functionality."""
    
    def test_plot_with_s2_data(self, dummy_pastis_data_full):
        """Test plotting with S2 RGB data."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=False,
            satellites=["S2"],
            pad_image=6,  # Ensure multiple rows for 2D subplot grid
        )
        sample = dataset[0]
        
        # Convert S2 to torch tensor for plot
        sample["S2"] = torch.from_numpy(sample["S2"])
        sample["target"] = torch.from_numpy(sample["mask"])
        
        fig = dataset.plot(sample, suptitle="Test Plot", show_axes=True)
        
        assert fig is not None
    
    def test_plot_without_s2_data(self, dummy_pastis_data_full):
        """Test plotting without S2 data returns None with warning."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2"],
        )
        sample = dataset[0]
        
        # Remove S2 from sample
        sample_no_s2 = {k: v for k, v in sample.items() if k != "S2"}
        sample_no_s2["target"] = torch.from_numpy(sample["mask"])
        
        with pytest.warns(UserWarning, match="No RGB image"):
            result = dataset.plot(sample_no_s2)
            assert result is None
    
    def test_plot_with_target(self, dummy_pastis_data_full):
        """Test plotting with target mask."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="semantic",
            satellites=["S2"],
            pad_image=6,  # Ensure multiple rows for 2D subplot grid
        )
        sample = dataset[0]
        
        # Convert to torch for plot function
        sample["S2"] = torch.from_numpy(sample["S2"])
        sample["target"] = torch.from_numpy(sample["mask"])
        
        fig = dataset.plot(sample, show_axes=False)
        assert fig is not None
    
    def test_plot_normalization(self, dummy_pastis_data_full):
        """Test that RGB normalization in plot handles edge cases."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2"],
            pad_image=6,  # Ensure multiple rows for 2D subplot grid
        )
        sample = dataset[0]
        
        # Create sample with constant RGB values (denom == 0 case)
        s2_data = sample["S2"]
        s2_data[:, :, :, 0] = 0.5  # Set all R channel to same value
        sample["S2"] = torch.from_numpy(s2_data)
        sample["target"] = torch.from_numpy(sample["mask"])
        
        # Should handle the case where denom == 0
        fig = dataset.plot(sample)
        assert fig is not None
    
    def test_plot_with_target_none(self, dummy_pastis_data_full):
        """Test plotting when target is None."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2"],
            pad_image=6,  # Ensure multiple rows for 2D subplot grid
        )
        sample = dataset[0]
        
        sample["S2"] = torch.from_numpy(sample["S2"])
        sample["target"] = None
        
        fig = dataset.plot(sample)
        assert fig is not None
    
    def test_plot_with_more_images_than_cols(self, dummy_pastis_data_full):
        """Test plotting with many timesteps to check grid layout."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2"],
            pad_image=10,  # Create 10 timesteps
        )
        sample = dataset[0]
        
        sample["S2"] = torch.from_numpy(sample["S2"])
        sample["target"] = torch.from_numpy(sample["mask"])
        
        fig = dataset.plot(sample)
        assert fig is not None


class TestPASTISEdgeCases:
    """Test edge cases and error conditions."""
    
    def test_len_method(self, dummy_pastis_data_full):
        """Test __len__ returns correct count."""
        dataset = PASTIS(data_root=dummy_pastis_data_full, folds=[1], satellites=["S2"])
        assert len(dataset) == 2
        
        dataset_all = PASTIS(data_root=dummy_pastis_data_full, satellites=["S2"])
        assert len(dataset_all) == 5
    
    def test_normalization_with_selected_folds(self, dummy_pastis_data_full):
        """Test that normalization uses correct fold statistics."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            folds=[1, 2],
            satellites=["S2"],
        )
        
        # Normalization should average stats from folds 1 and 2
        assert "S2" in dataset.norm
        assert len(dataset.norm["S2"]) == 2  # (mean, std)
    
    def test_multiple_satellites_different_channels(self, dummy_pastis_data_full):
        """Test handling satellites with different channel counts."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            satellites=["S2", "S1A"],
        )
        
        sample = dataset[0]
        
        # S2 has 3 channels, S1A has 2 channels
        assert sample["S2"].shape[1] == 3
        assert sample["S1A"].shape[1] == 2
    
    def test_date_table_construction(self, dummy_pastis_data_full):
        """Test that date tables are constructed correctly for all satellites."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2", "S1A", "S1D"],
        )
        
        assert "S2" in dataset.date_tables
        assert "S1A" in dataset.date_tables
        assert "S1D" in dataset.date_tables
        
        # Each patch should have date table entry
        for patch_id in [1, 2, 3, 4, 5]:
            assert patch_id in dataset.date_tables["S2"]
    
    def test_instance_semantic_annotation_consistency(self, dummy_pastis_data_full):
        """Test that instance target includes correct semantic annotations."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="instance",
            satellites=["S2"],
        )
        
        sample = dataset[0]
        mask = sample["mask"]
        
        # Check channel structure:
        # 0: heatmap, 1: instance_ids, 2: zones, 3-4: size, 5: object_semantic, 6: pixel_semantic
        assert mask.shape[-1] == 7
        
        # Pixel semantic should match the semantic target
        pixel_semantic = mask[..., 6]
        assert pixel_semantic.min() >= 0
        assert pixel_semantic.max() < 18
    
    def test_class_mapping_with_instance_target(self, dummy_pastis_data_full):
        """Test class mapping is applied to instance target annotations."""
        class_mapping = {i: i // 2 for i in range(18)}  # Group classes
        
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            target="instance",
            class_mapping=class_mapping,
            satellites=["S2"],
        )
        
        sample = dataset[0]
        mask = sample["mask"]
        
        # Pixel and object semantic annotations should use mapped classes
        pixel_semantic = mask[..., 6]
        object_semantic = mask[..., 5]
        
        # Check that mapping was applied
        unique_pixel = np.unique(pixel_semantic)
        unique_object = np.unique(object_semantic)
        
        assert len(unique_pixel) <= 9  # At most 9 mapped classes (18 // 2)


class TestPASTISIntegration:
    """Integration tests with realistic workflows."""
    
    def test_full_pipeline_semantic(self, dummy_pastis_data_full):
        """Test full pipeline for semantic segmentation."""
        def augmentation(**kwargs):
            # Simple augmentation
            if "image" in kwargs:
                kwargs["image"] = kwargs["image"] + 0.01
            return kwargs
        
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=True,
            target="semantic",
            folds=[1, 2],
            reference_date="2018-09-01",
            class_mapping={i: i % 10 for i in range(18)},
            transform=augmentation,
            truncate_image=5,
            pad_image=6,
            satellites=["S2"],
        )
        
        assert len(dataset) == 3
        sample = dataset[0]
        
        # Check all expected keys
        assert "image" in sample
        assert "mask" in sample
        assert "dates" in sample
        assert "S2" in sample
        
        # Check padding was applied
        assert sample["S2"].shape[0] == 6
    
    def test_full_pipeline_instance(self, dummy_pastis_data_full):
        """Test full pipeline for instance segmentation."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            norm=False,
            target="instance",
            folds=[1],
            satellites=["S2", "S1A"],
            truncate_image=3,
        )
        
        assert len(dataset) == 2
        sample = dataset[0]
        
        # Check instance-specific outputs
        assert sample["mask"].shape[-1] == 7
        assert "S2" in sample
        assert "S1A" in sample
    
    def test_iteration_over_dataset(self, dummy_pastis_data_full):
        """Test iterating over the full dataset."""
        dataset = PASTIS(
            data_root=dummy_pastis_data_full,
            satellites=["S2"],
        )
        
        samples = []
        for i in range(len(dataset)):
            sample = dataset[i]
            samples.append(sample)
        
        assert len(samples) == 5
        
        # Check all samples have required keys
        for sample in samples:
            assert "image" in sample
            assert "mask" in sample
            assert "dates" in sample



# Copyright contributors to the Terratorch project

import gc
import importlib

import pytest
import torch

from terratorch.models import EncoderDecoderFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS
from terratorch.models.model import AuxiliaryHead

NUM_CHANNELS = 6
NUM_CLASSES = 2
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)
EXPECTED_REGRESSION_OUTPUT_SHAPE = (1, 224, 224)
EXPECTED_CLASSIFICATION_OUTPUT_SHAPE = (1, NUM_CLASSES)

PIXELWISE_TASK_EXPECTED_OUTPUT = [
    ("regression", EXPECTED_REGRESSION_OUTPUT_SHAPE),
    ("segmentation", EXPECTED_SEGMENTATION_OUTPUT_SHAPE),
]

VIT_UPERNET_NECK = [
    {"name": "SelectIndices", "indices": [0, 1, 2, 3]},
    {"name": "ReshapeTokensToImage"},
    {"name": "LearnedInterpolateToPyramidal"},
]

LORA_CONFIG = {
    "prithvi": {
        "method": "LORA",
        "replace_qkv": "qkv",  # As we want to apply LoRA separately and only to Q and V, we need to separate the matrix.
        "peft_config_kwargs": {
            "target_modules": ["qkv.q_linear", "qkv.v_linear", "mlp.fc1", "mlp.fc2"],
            "lora_alpha": 16,
            "r": 16,
        },
    },
    "clay": {
        "method": "LORA",
        "replace_qkv": "to_qkv",
        "peft_config_kwargs": {
            "target_modules": ["to_qkv.q_linear", "to_qkv.v_linear", "1.net.1", "1.net.3"],
            "lora_alpha": 16,
            "r": 16,
        },
    },
}


@pytest.fixture(scope="session")
def model_factory() -> EncoderDecoderFactory:
    return EncoderDecoderFactory()


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))


def test_unused_args_raise_exception(model_factory: EncoderDecoderFactory):
    with pytest.raises(ValueError) as excinfo:
        model_factory.build_model(
            "classification",
            backbone="prithvi_eo_v1_100",
            decoder="IdentityDecoder",
            backbone_bands=PRETRAINED_BANDS,
            backbone_pretrained=False,
            num_classes=NUM_CLASSES,
            unused_argument="unused_argument",
        )
    assert "unused_argument" in str(excinfo.value)

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v2_300"])
def test_create_classification_model(backbone, model_factory: EncoderDecoderFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_bands=PRETRAINED_BANDS,
        backbone_pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v2_tiny_tl"])
def test_create_classification_model_no_in_channels(backbone, model_factory: EncoderDecoderFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_bands=PRETRAINED_BANDS,
        backbone_pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v2_100_tl"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["UNetDecoder"])
def test_create_pixelwise_model(backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input):
    if decoder == "LinearDecoder" and task == "regression":
        pytest.skip("LinearDecoder is not supported for regression tasks")
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    if decoder in ["UperNetDecoder", "UNetDecoder"] and backbone.startswith("prithvi_eo"):
        model_args["necks"] = VIT_UPERNET_NECK
    if decoder == "UNetDecoder":
        model_args["decoder_channels"] = [256, 128, 64, 32]
    if decoder == "LinearDecoder":
        model_args["decoder_upsampling_size"] = 16
        model_args["rescale"] = False

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
def test_create_model_with_smp_fpn_decoder(backbone, task, expected, model_factory: EncoderDecoderFactory, model_input):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": "smp_FPN",
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "necks": VIT_UPERNET_NECK,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
def test_create_model_with_smp_unet_decoder(
    backbone, task, expected, model_factory: EncoderDecoderFactory, model_input
):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": "smp_Unet",
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "decoder_decoder_channels": [256, 128, 64],
        "necks": VIT_UPERNET_NECK,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.skip(reason="Failing without clear reason.")
@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
def test_create_model_with_smp_deeplabv3plus_decoder(
    backbone, task, expected, model_factory: EncoderDecoderFactory, model_input
):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": "smp_DeepLabV3Plus",
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "necks": VIT_UPERNET_NECK + [{"name": "AddBottleneckLayer"}],
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == tuple(expected)

    gc.collect()


@pytest.mark.skipif(not importlib.util.find_spec("mmseg"), reason="mmsegmentation not installed")
@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
def test_create_model_with_mmseg_fcn_decoder(
    backbone, task, expected, model_factory: EncoderDecoderFactory, model_input
):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": "mmseg_FCNHead",
        "decoder_channels": 128,
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "necks": [
            {"name": "SelectIndices", "indices": [-1]},
            {"name": "ReshapeTokensToImage"},
        ],
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    else:
        model_args["num_classes"] = 1

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.skipif(not importlib.util.find_spec("mmseg"), reason="mmsegmentation not installed")
@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
def test_create_model_with_mmseg_uperhead_decoder(
    backbone, task, expected, model_factory: EncoderDecoderFactory, model_input
):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": "mmseg_UPerHead",
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "decoder_channels": 256,
        "decoder_in_index": [0, 1, 2, 3],
        "necks": [
            {"name": "SelectIndices", "indices": [0, 1, 2, 3]},
            {"name": "ReshapeTokensToImage"},
        ],
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    else:
        model_args["num_classes"] = 1

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v2_300_tl"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder"])
def test_create_pixelwise_model_no_in_channels(
    backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input
):
    if decoder == "LinearDecoder" and task == "regression":
        pytest.skip("LinearDecoder is not supported for regression tasks")
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    if decoder in ["UperNetDecoder", "UNetDecoder"] and backbone.startswith("prithvi_eo"):
        model_args["necks"] = VIT_UPERNET_NECK
    if decoder == "UNetDecoder":
        model_args["decoder_channels"] = [256, 128, 64, 32]
    if decoder == "LinearDecoder":
        model_args["decoder_upsampling_size"] = 16
        model_args["rescale"] = False

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["UperNetDecoder"])
def test_create_pixelwise_model_with_aux_heads(
    backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input
):
    aux_heads_name = ["first_aux", "second_aux"]
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_in_chans": NUM_CHANNELS,
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "aux_decoders": [AuxiliaryHead(name, "FCNDecoder", None) for name in aux_heads_name],
    }
    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    if decoder in ["UperNetDecoder", "UNetDecoder"] and backbone.startswith("prithvi_eo"):
        model_args["necks"] = VIT_UPERNET_NECK
    if decoder == "UNetDecoder":
        model_args["decoder_channels"] = [256, 128, 64, 32]

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        model_output = model(model_input)
        assert model_output.output.shape == expected

        assert len(model_output.auxiliary_heads.keys() & aux_heads_name) == len(aux_heads_name)
        for _, output in model_output.auxiliary_heads.items():
            assert output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder"])
def test_create_pixelwise_model_with_extra_bands(
    backbone, task, expected, decoder, model_factory: EncoderDecoderFactory
):
    if decoder == "LinearDecoder" and task == "regression":
        pytest.skip("LinearDecoder is not supported for regression tasks")
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_in_chans": NUM_CHANNELS + 1,
        "backbone_bands": [*PRETRAINED_BANDS, 7],
        "backbone_pretrained": False,
    }
    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    if decoder in ["UperNetDecoder", "UNetDecoder"] and backbone.startswith("prithvi_eo"):
        model_args["necks"] = VIT_UPERNET_NECK
    if decoder == "UNetDecoder":
        model_args["decoder_channels"] = [256, 128, 64, 32]
    if decoder == "LinearDecoder":
        model_args["decoder_upsampling_size"] = 16
        model_args["rescale"] = False
    model = model_factory.build_model(**model_args)
    model.eval()
    model_input = torch.ones((1, NUM_CHANNELS + 1, 224, 224))
    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100", "prithvi_eo_v2_300", "clay_v1_base"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["UNetDecoder"])
def test_create_model_with_lora(backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_bands": PRETRAINED_BANDS,
        "backbone_pretrained": False,
        "peft_config": LORA_CONFIG[backbone.split("_")[0]],
    }

    if backbone == "clay_v1_base":
        model_args["backbone_img_size"] = 224

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    if decoder in ["UperNetDecoder", "UNetDecoder"]:
        model_args["necks"] = VIT_UPERNET_NECK
    if decoder == "UNetDecoder":
        model_args["decoder_channels"] = [256, 128, 64, 32]

    model = model_factory.build_model(**model_args)
    model.eval()

    encoder_trainable_params = 0
    for param in model.encoder.parameters():
        if param.requires_grad:
            encoder_trainable_params += param.numel()

    if backbone == "clay_v1_base":
        num_layers = 12
        embed_dim = 768
        linear_in_out = 768 * 4
    else:
        num_layers = len(model.encoder.blocks)
        embed_dim = model.encoder.embed_dim
        linear_in_out = model.encoder.blocks[0].mlp.fc1.out_features
    r = LORA_CONFIG[backbone.split("_")[0]]["peft_config_kwargs"]["r"]
    assert encoder_trainable_params == num_layers * (2 * embed_dim * r * 2 + 2 * linear_in_out * r + 2 * embed_dim * r)

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()



import pytest
import torch
import lightning
from terratorch.models import SMPModelFactory
from torch import nn
from terratorch.tasks.optimizer_factory import optimizer_factory


NUM_CHANNELS = 3
MODEL_TYPE = "Unet"
NUM_CLASSES = 2
PRETRAINED_BANDS = ["RED", "GREEN", "BLUE"]
PRETRAINED=False
BACKBONE="resnet50"
TASK_TYPE="segmentation"
IMAGE_SIZE=224
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, IMAGE_SIZE, IMAGE_SIZE)
DEFAULT_INPUTS = {
    "optimizer": "AdamW",
    "optimizer_hparams": {
        "weight_decay": 0.01,
        },
    "monitor": "val_loss",
    "lr": 0.001
    }

WARMUP_DECAY_INPUTS = [
    (
        {"schedulers":{
                "LambdaLR": {"lr_lambda": "linear_warmup"},
                "CosineAnnealingLR": {"T_max": 10}
                },
        "milestones": [5],
        }, DEFAULT_INPUTS, 10
    ),
    (
        {"schedulers":{
                "LambdaLR": {"lr_lambda": "linear_warmup"},
                "ExponentialLR": {"gamma":0.95}
                },
        "milestones": [2],
        },DEFAULT_INPUTS, 5
    )
]
WARMUP_DECAY_TEST_IDS = [str(i) for i in range(0, len(WARMUP_DECAY_INPUTS))]



@pytest.fixture(scope="session")
def model() -> nn.Module:
    model_factory = SMPModelFactory()
    model = model_factory.build_model(
        task=TASK_TYPE,
        backbone=BACKBONE,
        model=MODEL_TYPE,
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=PRETRAINED,
        num_classes=NUM_CLASSES,
    )
    return model

@pytest.mark.parametrize(
    "scheduler_hparams, defaults, epochs", 
    WARMUP_DECAY_INPUTS,
    ids=WARMUP_DECAY_TEST_IDS,
)
def test_warmup_sequential_scheduling(
    scheduler_hparams:dict,
    defaults: str,
    epochs: int,
    model: nn.Module,
    ):
    """ Test sequentiallr scheduler with warm up stage then decay stage
    """
    model.eval()
    loss_fn = nn.CrossEntropyLoss()

    defaults.update({"scheduler": "SequentialLR"})
    defaults.update({"scheduler_hparams":scheduler_hparams})
    defaults.update({"params_to_be_optimized": model.parameters()})
    optimizer_factory_output = optimizer_factory(**defaults)

    #check type
    assert isinstance(optimizer_factory_output["optimizer"], torch.optim.Optimizer)
    assert isinstance(optimizer_factory_output["lr_scheduler"]["scheduler"], torch.optim.lr_scheduler.SequentialLR )

    scheduler = optimizer_factory_output["lr_scheduler"]["scheduler"]
    optimizer = optimizer_factory_output["optimizer"]

    def training_step(optimizer, scheduler, model, loss_fn):
        optimizer.zero_grad()
        model_input = torch.rand((1, len(PRETRAINED_BANDS), IMAGE_SIZE, IMAGE_SIZE))
        target = torch.ones(EXPECTED_SEGMENTATION_OUTPUT_SHAPE)
        model_output = model(model_input)
        loss = loss_fn(model_output.output, target)
        loss.backward()
        optimizer.step()
        scheduler.step()
    
    #check that learning rate is changing according to schedule
    for i in range(epochs):
        pre_step_lr = scheduler.get_last_lr()
        training_step(optimizer, scheduler, model, loss_fn)
        post_step_lr = scheduler.get_last_lr()
        if i < scheduler_hparams["milestones"][0]-1:
            assert pre_step_lr < post_step_lr, "LR is not increasing during warmup stage"
        elif i == scheduler_hparams["milestones"][0]-1:
            assert pre_step_lr == post_step_lr, "LR does not match expected LR at this point"
        else:
            assert pre_step_lr > post_step_lr, "LR is not decreasing during decay stage"

#add tests for repeated (cyclic) warmup and decay

#add parametrized tests with other oprimizers



from lightning.pytorch.callbacks import Callback
from lightning.pytorch.callbacks.finetuning import BaseFinetuning


class ProgressiveResizing(Callback):
    def __init__(self):
        self.resize_schedule = {
            0: {"batch_size": 4, "num_workers": 4, "size": 64},
            10: {"batch_size": 2, "num_workers": 2, "size": 128},
            20: {"batch_size": 1, "num_workers": 1, "size": 256},
        }

    def on_train_epoch_start(self, trainer, pl_module):
        if trainer.current_epoch in self.resize_schedule:
            params = self.resize_schedule[trainer.current_epoch]

            trainer.datamodule.size = params["size"]
            trainer.datamodule.batch_size = params["batch_size"]
            trainer.datamodule.num_workers = params["num_workers"]

            trainer.datamodule.setup(stage="fit")

    def on_validation_epoch_start(self, trainer, pl_module):
        if trainer.current_epoch in self.resize_schedule:
            params = self.resize_schedule[trainer.current_epoch]

            trainer.datamodule.size = params["size"]
            trainer.datamodule.batch_size = params["batch_size"]
            trainer.datamodule.num_workers = params["num_workers"]

            trainer.datamodule.setup(stage="validate")


class LayerwiseFinetuning(BaseFinetuning):
    def __init__(self, phase, train_bn=True):
        """Initializes with phase & batch-norm information.

        Args:
            phase (List): Phases of fine-tuning the backbone network.
            train_bn (bool, optional): Trains just the batch-norm layers even
            when all the other layers of the network are freezed. Defaults to True.
        """
        super().__init__()
        self.phase = phase
        self.train_bn = train_bn

    def freeze_before_training(self, pl_module):
        """Freezes the encoder before starting the training."""
        self.freeze(
            modules=[
                pl_module.model.encoder.patch_embedding,
                pl_module.model.encoder.transformer,
            ],
            train_bn=self.train_bn,
        )

    def finetune_function(self, pl_module, epoch, optimizer):
        if epoch == self.phase:
            """Unfreezes the encoder for training."""
            print(f"In Phase {self.phase}: Full throttle")
            self.unfreeze_and_add_param_group(
                modules=[
                    pl_module.model.encoder.patch_embedding,
                    pl_module.model.encoder.transformer,
                ],
                optimizer=optimizer,
                train_bn=self.train_bn,
            )
            params = list(pl_module.parameters())
            active = list(filter(lambda p: p.requires_grad, params))
            print(f"active: {len(active)}, all: {len(params)}")



"""Ugly proxy objects so parsing config file works with transforms.

    These are necessary since, for LightningCLI to instantiate arguments as
    objects from the config, they must have type annotations

    In TorchGeo, `transforms` is passed in **kwargs, so it has no type annotations!
    To get around that, we create these wrappers that have transforms type annotated.
    They create the transforms and forward all method and attribute calls to the
    original TorchGeo datamodule.

    Additionally, TorchGeo datasets pass the data to the transforms callable
    as a dict, and as a tensor.

    Albumentations expects this data not as a dict but as different key-value
    arguments, and as numpy. We handle that conversion here. 
"""
from collections.abc import Callable
from typing import Any, Union
from functools import partial
import numpy as np
from torch import Tensor
from torch.utils.data import default_collate
import kornia.augmentation as K
from kornia.augmentation._2d.geometric.base import GeometricAugmentationBase2D
from kornia.augmentation._2d.intensity.base import IntensityAugmentationBase2D
import torch.nn as nn
from terratorch.datasets.transforms import kornia_augmentations_to_callable_with_dict
import torch
from einops import rearrange
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from torchgeo.datasets.utils import percentile_normalization, lazy_import
from matplotlib import patches
from geobench_v2.datamodules import (
    GeoBenchClassificationDataModule,
    GeoBenchSegmentationDataModule,
    GeoBenchObjectDetectionDataModule,
    MultiTemporalSegmentationAugmentation,
    MultiModalSegmentationAugmentation,
    )


class GeoBenchV2ClassificationDataModule(GeoBenchClassificationDataModule):
    """Proxy object for using Classification DataModules defined by geobench_v2.

    Allows for transforms to be defined and passed using config files.
    The only reason this class exists is so that we can annotate the transforms argument with a type.
    This is required for lightningcli and config files.
    As such, all getattr and setattr will be redirected to the underlying class.
    """

    def __init__(
        self,
        cls: type[GeoBenchClassificationDataModule],
        img_size: int,
        band_order: list | dict,
        batch_size: int | None = None,
        num_workers: int = 0,
        train_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]] | str = "default",
        eval_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]] | str = "default",
        predict_loader: str = 'test',
        **kwargs: Any,
    ):
        """Constructor

        Args:
            cls (type[GeoBenchClassificationDataModule]): geobench_v2 Classification DataModule class to be instantiated
            batch_size (int | None, optional): batch_size. Defaults to None.
            num_workers (int, optional): num_workers. Defaults to 0.
            transforms (None | list[Union[GeometricAugmentationBase2D, K.VideoSequential]], optional): List of Albumentations Transforms.
                Should enc with ToTensorV2. Defaults to None.
            predict_loader (str, optional): Selects whether predictions run on the train, val, or test split (default: "test").
            **kwargs (Any): Arguments passed to instantiate `cls`.
        """
        if isinstance(train_augmentations, str):
            msg = "If train_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert train_augmentations in ["default", "multi_temporal_default"], msg
            
        if isinstance(eval_augmentations, str):
            msg = "If eval_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert eval_augmentations in ["default", "multi_temporal_default"], msg

        kwargs["img_size"] = img_size
        kwargs["band_order"] = band_order

        if batch_size is not None:
            kwargs["batch_size"] = batch_size

        if not train_augmentations in [None, "default", "multi_temporal_default"]:
            train_augmentations = kornia_augmentations_to_callable_with_dict(train_augmentations)
        kwargs["train_augmentations"] = train_augmentations
            
        if not eval_augmentations in [None, "default", "multi_temporal_default"]:
            eval_augmentations = kornia_augmentations_to_callable_with_dict(eval_augmentations)
        kwargs["eval_augmentations"] = eval_augmentations
        
        if isinstance(band_order, list): 
            if len(band_order) > 0:
                if isinstance(band_order[0], dict):
                    band_order_dict = {}
                    for modality in band_order:
                        band_order_dict.update(modality)
                    band_order = band_order_dict
                    kwargs["band_order"] = band_order

        self._proxy = cls(num_workers=num_workers, **kwargs)
        self.predict_loader = predict_loader
        super().__init__(
            dataset_class = self._proxy.dataset_class, 
            img_size = img_size,
            band_order = band_order)

    @property
    def collate_fn(self):
        return self._proxy.collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._proxy.collate_fn = value

    def setup(self, stage: str):
        return self._proxy.setup(stage)

    def train_dataloader(self):
        return self._proxy.train_dataloader()

    def val_dataloader(self):
        return self._proxy.val_dataloader()

    def test_dataloader(self):
        return self._proxy.test_dataloader()

    def predict_dataloader(self):
        if self.predict_loader == "train":
            return self._proxy.train_dataloader()
        elif self.predict_loader == "val":
            return self._proxy.val_dataloader()
        elif self.predict_loader == "test":
            return self._proxy.test_dataloader()

    def _valid_attribute(self, *args: str):
        return self._proxy._valid_attribute(args)

    def plot(self, batch, split: str = "val"):
        return self._proxy.visualize_batch(batch = batch, split = split)




class GeoBenchV2ObjectDetectionDataModule(GeoBenchObjectDetectionDataModule):
    """Proxy object for using Object Detection DataModules defined by geobench_v2.

    Allows for transforms to be defined and passed using config files.
    The only reason this class exists is so that we can annotate the transforms argument with a type.
    This is required for lightningcli and config files.
    As such, all getattr and setattr will be redirected to the underlying class.
    """

    def __init__(
        self,
        cls: type[GeoBenchObjectDetectionDataModule],
        root: str,
        img_size: int,
        band_order: Any,
        categories: list,
        batch_size: int | None = None,
        eval_batch_size: int | None = None,
        num_workers: int = 0,
        train_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]]| str = "default",
        eval_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]] | str = "default",
        plot_indexes: list = [0,1,2],
        collate_fn: Callable = None,
        predict_loader: str = 'test',
        **kwargs: Any,
    ):
        """Constructor

        Args:
            cls (type[GeoBenchObjectDetectionDataModule]): geobench_v2 Object Detection DataModule class to be instantiated
            batch_size (int | None, optional): batch_size. Defaults to None.
            num_workers (int, optional): num_workers. Defaults to 0.
            transforms (None | list[Union[GeometricAugmentationBase2D, K.VideoSequential]], optional): List of Albumentations Transforms.
                Should enc with ToTensorV2. Defaults to None.
            predict_loader (str, optional): Selects whether predictions run on the train, val, or test split (default: "test").
            **kwargs (Any): Arguments passed to instantiate `cls`.
        """
        if isinstance(train_augmentations, str):
            msg = "If train_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert train_augmentations in ["default", "multi_temporal_default"], msg
            
        if isinstance(eval_augmentations, str):
            msg = "If eval_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert eval_augmentations in ["default", "multi_temporal_default"], msg

        kwargs['root'] = root
        kwargs["img_size"] = img_size
        kwargs["band_order"] = band_order

        if batch_size is not None:
            kwargs["batch_size"] = batch_size
        if eval_batch_size is not None:
            kwargs["eval_batch_size"] = eval_batch_size
        if not train_augmentations in [None, "default", "multi_temporal_default"]:
            train_augmentations = kornia_augmentations_to_callable_with_dict(train_augmentations)
        kwargs["train_augmentations"] = train_augmentations
            
        if not eval_augmentations in [None, "default", "multi_temporal_default"]:
            eval_augmentations = kornia_augmentations_to_callable_with_dict(eval_augmentations)
        kwargs["eval_augmentations"] = eval_augmentations
        
        if isinstance(band_order, list): 
            if len(band_order) > 0:
                if isinstance(band_order[0], dict):
                    band_order_dict = {}
                    for modality in band_order:
                        band_order_dict.update(modality)
                    band_order = band_order_dict
                    kwargs["band_order"] = band_order

        self._proxy = cls(num_workers=num_workers, **kwargs)
        self.predict_loader = predict_loader
        super().__init__(
            dataset_class = self._proxy.dataset_class, 
            img_size = img_size,
            band_order = band_order,
            batch_size = batch_size,
            eval_batch_size = eval_batch_size)  # dummy arg
        
        self.collate_fn = collate_fn
        self.plot_indexes = plot_indexes
        self.categories = categories

    @property
    def collate_fn(self):
        return self._proxy.collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._proxy.collate_fn = value

    @property
    def patch_size(self):
        return self._proxy.patch_size

    @property
    def length(self):
        return self._proxy.length

    def setup(self, stage: str):
        return self._proxy.setup(stage)

    def train_dataloader(self):
        return self._proxy.train_dataloader()

    def val_dataloader(self):
        return self._proxy.val_dataloader()

    def test_dataloader(self):
        return self._proxy.test_dataloader()

    def predict_dataloader(self):
        if self.predict_loader == "train":
            return self._proxy.train_dataloader()
        elif self.predict_loader == "val":
            return self._proxy.val_dataloader()
        elif self.predict_loader == "test":
            return self._proxy.test_dataloader()

    def _valid_attribute(self, *args: str):
        return self._proxy._valid_attribute(args)
    
    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        show_feats: str | None = 'both',
        box_alpha: float = 0.7,
        mask_alpha: float = 0.7,
        confidence_score = 0.5
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle
            show_titles: flag indicating whether to show titles above each panel
            show_feats: optional string to pick features to be shown: boxes, masks, both
            box_alpha: alpha value of box
            mask_alpha: alpha value of mask

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            AssertionError: if ``show_feats`` argument is invalid
            DependencyNotFoundError: If plotting masks and scikit-image is not installed.

        .. versionadded:: 0.4
        """
        assert show_feats in {'boxes', 'masks', 'both'}

        image = sample['image']
        image = image.median(1).values if len(image.shape) == 4 else image
        # get indexes to plot 
        image = image[ self.plot_indexes, :, :]
        if image.mean() > 1:
            image = image / 10000
        
        image = percentile_normalization(image.permute(1, 2, 0).numpy())

        if show_feats != 'boxes':
            skimage = lazy_import('skimage')

        boxes = sample['boxes'].cpu().numpy()
        labels = sample['labels'].cpu().numpy()

        if 'masks' in sample:
            masks = [mask.squeeze().cpu().numpy() for mask in sample['masks']]

        n_gt = len(boxes)

        ncols = 1
        show_predictions = 'prediction_labels' in sample

        if show_predictions:
            show_pred_boxes = False
            show_pred_masks = False
            prediction_labels = sample['prediction_labels'].numpy()
            prediction_scores = sample['prediction_scores'].numpy()
            if 'prediction_boxes' in sample:
                prediction_boxes = sample['prediction_boxes'].numpy()
                show_pred_boxes = True
            if 'prediction_masks' in sample:
                prediction_masks = sample['prediction_masks'].numpy()
                show_pred_masks = True

            n_pred = len(prediction_labels)
            ncols += 1

        # Display image
        fig, axs = plt.subplots(ncols=ncols, squeeze=False, figsize=(ncols * 10, 13))
        axs[0, 0].imshow(image)
        axs[0, 0].axis('off')

        cm = plt.get_cmap('gist_rainbow')
        for i in range(n_gt):
            class_num = labels[i]
            color = cm(class_num / len(self.categories))

            # Add bounding boxes
            x1, y1, x2, y2 = boxes[i]
            if show_feats in {'boxes', 'both'}:
                r = patches.Rectangle(
                    (x1, y1),
                    x2 - x1,
                    y2 - y1,
                    linewidth=2,
                    alpha=box_alpha,
                    linestyle='dashed',
                    edgecolor=color,
                    facecolor='none',
                )
                axs[0, 0].add_patch(r)

            # Add labels
            label = self.categories[class_num]
            caption = label
            axs[0, 0].text(
                x1, y1 - 8, caption, color='white', size=11, backgroundcolor='none'
            )

            # Add masks
            if show_feats in {'masks', 'both'} and 'masks' in sample:
                mask = masks[i]
                contours = skimage.measure.find_contours(mask, 0.5)
                for verts in contours:
                    verts = np.fliplr(verts)
                    p = patches.Polygon(
                        verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                    )
                    axs[0, 0].add_patch(p)

            if show_titles:
                axs[0, 0].set_title('Ground Truth')

        if show_predictions:
            axs[0, 1].imshow(image)
            axs[0, 1].axis('off')
            for i in range(n_pred):
                score = prediction_scores[i]
                if score < confidence_score:
                    continue

                class_num = prediction_labels[i]
                color = cm(class_num / len(self.categories))

                if show_pred_boxes:
                    # Add bounding boxes
                    x1, y1, x2, y2 = prediction_boxes[i]
                    r = patches.Rectangle(
                        (x1, y1),
                        x2 - x1,
                        y2 - y1,
                        linewidth=2,
                        alpha=box_alpha,
                        linestyle='dashed',
                        edgecolor=color,
                        facecolor='none',
                    )
                    axs[0, 1].add_patch(r)

                    # Add labels
                    label = self.categories[class_num]
                    caption = f'{label} {score:.3f}'
                    axs[0, 1].text(
                        x1,
                        y1 - 8,
                        caption,
                        color='white',
                        size=11,
                        backgroundcolor='none',
                    )

                # Add masks
                if show_pred_masks:

                    mask = prediction_masks[i][0]
                    contours = skimage.measure.find_contours(mask, 0.5)
                    for verts in contours:
                        verts = np.fliplr(verts)
                        p = patches.Polygon(
                            verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                        )
                        axs[0, 1].add_patch(p)

            if show_titles:
                axs[0, 1].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        plt.tight_layout()

        return fig




class GeoBenchV2SegmentationDataModule(GeoBenchSegmentationDataModule):
    """Proxy object for using Segmentation DataModules defined by geobench_v2.

    Allows for transforms to be defined and passed using config files.
    The only reason this class exists is so that we can annotate the transforms argument with a type.
    This is required for lightningcli and config files.
    As such, all getattr and setattr will be redirected to the underlying class.
    """

    def __init__(
        self,
        cls: type[GeoBenchSegmentationDataModule],
        img_size: int,
        band_order: list,
        batch_size: int | None = None,
        num_workers: int = 0,
        train_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]] | str = "default",
        eval_augmentations: None | list[Union[GeometricAugmentationBase2D, K.VideoSequential, IntensityAugmentationBase2D]] | str = "default",
        collate_fn: Callable = None,
        predict_loader: str = 'test',
        **kwargs: Any,
    ):
        """Constructor

        Args:
            cls (type[GeoDataModule]): geobench_v2 Segmentation DataModule class to be instantiated
            batch_size (int | None, optional): batch_size. Defaults to None.
            num_workers (int, optional): num_workers. Defaults to 0.
            transforms (None | list[Union[GeometricAugmentationBase2D, K.VideoSequential]], optional): List of Kornia Transforms.
                Should enc with ToTensorV2. Defaults to None.
            predict_loader (str, optional): Selects whether predictions run on the train, val, or test split (default: "test").
            **kwargs (Any): Arguments passed to instantiate `cls`.
        """
        if isinstance(train_augmentations, str):
            msg = "If train_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert train_augmentations in ["default", "multi_temporal_default"], msg
            
        if isinstance(eval_augmentations, str):
            msg = "If eval_augmentations is a string, it must be 'default' or 'multi_temporal_default'"
            assert eval_augmentations in ["default", "multi_temporal_default"], msg
            
        kwargs["img_size"] = img_size
        kwargs["band_order"] = band_order
        
        if batch_size is not None:
            kwargs["batch_size"] = batch_size

        if "rename_modalities" in kwargs:
            if train_augmentations == "default":
                train_augmentations = [ K.RandomHorizontalFlip(p=0.5), K.RandomVerticalFlip(p=0.5)]
            elif train_augmentations == "multi_temporal_default":
                train_augmentations = [K.VideoSequential(), K.RandomHorizontalFlip(p=0.5), K.RandomVerticalFlip(p=0.5)]
            elif train_augmentations is None:
                train_augmentations = [nn.Identity()]
            train_augmentations = kornia_augmentations_to_callable_with_dict(train_augmentations)
            train_augmentations = MultiModalSegmentationAugmentation(transforms=train_augmentations)

            if eval_augmentations in ["default", None]:
                eval_augmentations = [nn.Identity()]
            elif eval_augmentations == "multi_temporal_default":
                eval_augmentations = [K.VideoSequential(), nn.Identity()]
            eval_augmentations = kornia_augmentations_to_callable_with_dict(eval_augmentations)
            eval_augmentations = MultiModalSegmentationAugmentation(transforms=eval_augmentations)
        else:
            if not train_augmentations in [None, "default", "multi_temporal_default"]:
                if isinstance(train_augmentations[0], K.VideoSequential):
                    train_augmentations = kornia_augmentations_to_callable_with_dict(train_augmentations)
                    train_augmentations = MultiTemporalSegmentationAugmentation(transforms=train_augmentations)
                else:
                    train_augmentations = kornia_augmentations_to_callable_with_dict(train_augmentations)
            
            if not eval_augmentations in [None, "default", "multi_temporal_default"]:
                if isinstance(eval_augmentations[0], K.VideoSequential):
                    eval_augmentations = kornia_augmentations_to_callable_with_dict(eval_augmentations)
                    eval_augmentations = MultiTemporalSegmentationAugmentation(transforms=eval_augmentations)
                else:
                    eval_augmentations = kornia_augmentations_to_callable_with_dict(eval_augmentations)

        kwargs["train_augmentations"] = train_augmentations
        kwargs["eval_augmentations"] = eval_augmentations

        if len(band_order) > 0:
            if isinstance(band_order[0], dict):
                band_order_dict = {}
                for modality in band_order:
                    band_order_dict.update(modality)
                band_order = band_order_dict
                kwargs["band_order"] = band_order

        self._proxy = cls(num_workers=num_workers, **kwargs)
        self.predict_loader = predict_loader

        super().__init__(
            dataset_class = self._proxy.dataset_class, 
            img_size = img_size,
            band_order = band_order
            )  # dummy arg


        self.collate_fn = collate_fn

    @property
    def collate_fn(self):
        return self._proxy.collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._proxy.collate_fn = value

    @property
    def patch_size(self):
        return self._proxy.patch_size

    @property
    def length(self):
        return self._proxy.length

    @property
    def val_dataset(self):
        return self._proxy.val_dataset

    def setup(self, stage: str):
        return self._proxy.setup(stage)

    def train_dataloader(self):
        return self._proxy.train_dataloader()

    def val_dataloader(self):
        return self._proxy.val_dataloader()

    def test_dataloader(self):
        return self._proxy.test_dataloader()

    def predict_dataloader(self):
        if self.predict_loader == "train":
            return self._proxy.train_dataloader()
        elif self.predict_loader == "val":
            return self._proxy.val_dataloader()
        elif self.predict_loader == "test":
            return self._proxy.test_dataloader()

    def _valid_attribute(self, *args: str):
        return self._proxy._valid_attribute(args)

    def plot(
        self, batch: dict[str, Any] | None = None, split: str = "train"
    ) -> tuple[Any, dict[str, Any]]:
        """Visualize a batch of data with output.

        Args:
            batch: Optional batch of data. If not provided, a batch will be fetched from the dataloader.
            split: One of 'train', 'validation', 'test'

        Returns:
            The matplotlib figure and the batch of data
        """
        if batch is None:
            if split == "train":
                batch = next(iter(self.train_dataloader()))
            elif split == "validation":
                batch = next(iter(self.val_dataloader()))
            else:
                batch = next(iter(self.test_dataloader()))

        if hasattr(self._proxy.data_normalizer, "unnormalize"):
            batch = self._proxy.data_normalizer.unnormalize(batch)

        try:
            masks = batch["mask"]
            images = batch["image"]
            if isinstance(images, dict):
                for k in images:
                    print(f'{k=} {v.shape=}')
                images = next(iter(images.values())) 
            prediction = batch["prediction"] if "prediction" in batch else None
            if prediction is not None:
                num_cols = 3
            else:
                num_cols = 2
    
            n_samples = images.shape[0]
            n_samples = min(3, n_samples)
            indices = torch.randperm(n_samples)[:n_samples]
    
            images = images[indices]
            masks = masks[indices] if masks.dim()>2 else masks
    
            plot_bands = self._proxy.dataset_band_config.plot_bands
            rgb_indices = [0,1,2]
            if images.dim() == 4:
                images = images[:, rgb_indices, :, :]
            elif images.dim() == 3:
                images = images[rgb_indices, :, :]
            else: #if multi-temporal choose first timestamp
                images = images[:, rgb_indices, 0, :, :]
    
            fig, axes = plt.subplots(n_samples, num_cols, figsize=(12, 3 * n_samples))
            if n_samples == 1:
                axes = axes.reshape(1, -1)
    
            for i in range(n_samples):
                ax = axes[i, 0]
                img = images[i] if images.dim()>3 else images
                img = rearrange(img, "c h w -> h w c").cpu().numpy()
                img = percentile_normalization(img, lower=2, upper=98)
                ax.imshow(img)
                ax.set_title("image")
                ax.axis("off")
    
                ax = axes[i, 1]
                mask_img = masks[i] if masks.dim()>2 else masks
                mask_img = mask_img.cpu().numpy()
                ax.imshow(mask_img)
                ax.set_title("mask" if i == 0 else "")
                ax.axis("off")
    
                if prediction is not None:
                    ax = axes[i, 2]
                    prediction_img = prediction[i] if masks.dim()>2 else prediction
                    prediction_img = prediction_img.cpu().numpy()
                    ax.imshow(prediction_img)
                    ax.set_title("prediction" if i == 0 else "")
                    ax.axis("off")
    
            plt.tight_layout()
            return fig
        except Exception as e:
            print(e)
            return None


from collections.abc import Sequence
from typing import Any

import albumentations as A
from kornia.augmentation import AugmentationSequential

from terratorch.datamodules.geobench_data_module import GeobenchDataModule
from terratorch.datasets import MPv4gerSegNonGeo

MEANS = {"BLUE": 139.761751, "GREEN": 137.354091, "RED": 131.102356}

STDS = {"BLUE": 48.29800594656056, "GREEN": 50.86544377633718, "RED": 54.52768048660482}


class MPv4gerSegNonGeoDataModule(GeobenchDataModule):
    """NonGeo LightningDataModule implementation for M-Pv4gerSeg dataset."""

    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        bands: Sequence[str] | None = None,
        train_transform: A.Compose | None | list = None,
        val_transform: A.Compose | None | list = None,
        test_transform: A.Compose | None | list = None,
        aug: AugmentationSequential = None,
        partition: str = "default",
        use_metadata: bool = False,  # noqa: FBT002, FBT001
        **kwargs: Any,
    ) -> None:
        """
        Initializes the MPv4gerNonGeoDataModule for the MPv4gerSegNonGeo dataset.

        Args:
            batch_size (int, optional): Batch size for DataLoaders. Defaults to 8.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            data_root (str, optional): Root directory of the dataset. Defaults to "./".
            bands (Sequence[str] | None, optional): List of bands to use. Defaults to None.
            train_transform (A.Compose | None | list, optional): Transformations for training.
            val_transform (A.Compose | None | list, optional): Transformations for validation.
            test_transform (A.Compose | None | list, optional): Transformations for testing.
            aug (AugmentationSequential, optional): Augmentation/normalization pipeline. Defaults to None.
            partition (str, optional): Partition size. Defaults to "default".
            use_metadata (bool): Whether to return metadata info.
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(
            MPv4gerSegNonGeo,
            MEANS,
            STDS,
            batch_size=batch_size,
            num_workers=num_workers,
            data_root=data_root,
            bands=bands,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            aug=aug,
            partition=partition,
            use_metadata=use_metadata,
            **kwargs,
        )



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import logging
import warnings
from torch import nn
from functools import partial
from .terramind import TerraMind
from .terramind_vit import TerraMindViT
from .terramind_tim import TerraMindTiM
from .terramind_generation import TerraMindGeneration
from .tm_utils import LayerNorm
from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY, TERRATORCH_FULL_MODEL_REGISTRY
from huggingface_hub import hf_hub_download

from terratorch.models.backbones.terramind.tokenizer.tokenizer_register import (
    terramind_v1_tokenizer_s2l2a,
    terramind_v1_tokenizer_s1rtc,
    terramind_v1_tokenizer_s1grd,
    terramind_v1_tokenizer_dem,
    terramind_v1_tokenizer_lulc,
    terramind_v1_tokenizer_ndvi,
    terramind_v01_tokenizer_s2l2a,
    terramind_v01_tokenizer_s1grd,
    terramind_v01_tokenizer_dem,
    terramind_v01_tokenizer_lulc,
    terramind_v01_caption_tokenizer,
    terramind_v1_coords_tokenizer,
)

logger = logging.getLogger('terramind')


# Model definitions
__all__ = [
    # pre-trained models
    "terramind_v01_base",
    "terramind_v1_tiny",
    "terramind_v1_small",
    "terramind_v1_base",
    "terramind_v1_large",
    "terramind_v1_tiny_tim",
    "terramind_v1_small_tim",
    "terramind_v1_base_tim",
    "terramind_v1_large_tim",
    "terramind_v1_tiny_encdec",
    "terramind_v1_small_encdec",
    "terramind_v1_base_encdec",
    "terramind_v1_large_encdec",
    "terramind_v1_tiny_generate",
    "terramind_v1_small_generate",
    "terramind_v1_base_generate",
    "terramind_v1_large_generate",
]

pretrained_weights = {
        "terramind_v01_base": {
            "hf_hub_id": "FAST-EO/TerraMind-0.1-base",
            "hf_hub_filename": "TerraMind_v01_base.pt",
        },
        "terramind_v1_tiny": {
            "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-tiny",
            "hf_hub_filename": "TerraMind_v1_tiny.pt",
        },
        "terramind_v1_small": {
            "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-small",
            "hf_hub_filename": "TerraMind_v1_small.pt",
        },
        "terramind_v1_base": {
            "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-base",
            "hf_hub_filename": "TerraMind_v1_base.pt",
        },
        "terramind_v1_large": {
            "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-large",
            "hf_hub_filename": "TerraMind_v1_large.pt",
        },
    }

PRETRAINED_BANDS = {
    "untok_sen2l2a@224": [
        "COASTAL_AEROSOL",
        "BLUE",
        "GREEN",
        "RED",
        "RED_EDGE_1",
        "RED_EDGE_2",
        "RED_EDGE_3",
        "NIR_BROAD",
        "NIR_NARROW",
        "WATER_VAPOR",
        "SWIR_1",
        "SWIR_2",
    ],
    "untok_sen2l1c@224": [
        "COASTAL_AEROSOL",
        "BLUE",
        "GREEN",
        "RED",
        "RED_EDGE_1",
        "RED_EDGE_2",
        "RED_EDGE_3",
        "NIR_BROAD",
        "NIR_NARROW",
        "WATER_VAPOR",
        "CIRRUS",
        "SWIR_1",
        "SWIR_2",
    ],
    "untok_sen2rgb@224": ["RED", "GREEN", "BLUE"],
    "untok_sen1grd@224": ["VV", "VH"],
    "untok_sen1rtc@224": ["VV", "VH"],
    "untok_dem@224": ["DEM"],
}


v01_pretraining_mean = {
    "untok_sen2l2a@224": [794.311,  925.161, 1183.128, 1338.041, 1667.254, 2233.633, 2460.96 , 2555.569, 2619.542, 2703.298, 2406.497, 1841.645],
    "tok_sen2l2a@224": [794.311,  925.161, 1183.128, 1338.041, 1667.254, 2233.633, 2460.96 , 2555.569, 2619.542, 2703.298, 2406.497, 1841.645],
    "tok_sen1grd@224": [-12.599, -20.293],
    "tok_lulc@224": [0],
    "tok_dem@224": [435.726],
}

v01_pretraining_std = {
    "untok_sen2l2a@224": [1164.883, 1205.586, 1223.713, 1399.638, 1403.298, 1378.513, 1434.924, 1491.141, 1454.089, 1660.395, 1473.248, 1365.080],
    "tok_sen2l2a@224": [1164.883, 1205.586, 1223.713, 1399.638, 1403.298, 1378.513, 1434.924, 1491.141, 1454.089, 1660.395, 1473.248, 1365.080],
    "tok_sen1grd@224": [5.195, 5.890],
    "tok_lulc@224": [1],
    "tok_dem@224": [560.326],
}

v1_pretraining_mean = {
    "untok_sen2l2a@224": [1390.458, 1503.317, 1718.197, 1853.91, 2199.1, 2779.975, 2987.011, 3083.234, 3132.22, 3162.988, 2424.884, 1857.648],
    "untok_sen2l1c@224": [2357.089, 2137.385, 2018.788, 2082.986, 2295.651, 2854.537, 3122.849, 3040.56, 3306.481, 1473.847,  506.07, 2472.825, 1838.929],
    "untok_sen2rgb@224": [87.271, 80.931, 66.667],
    "untok_sen1grd@224": [-12.599, -20.293],
    "untok_sen1rtc@224": [-10.93, -17.329],
    "untok_dem@224": [670.665],
    "tok_sen1grd@224": [-12.599, -20.293],
    "tok_sen1rtc@224": [-10.93, -17.329],
    "tok_sen2l2a@224": [1390.458, 1503.317, 1718.197, 1853.91, 2199.1, 2779.975, 2987.011, 3083.234, 3132.22, 3162.988, 2424.884, 1857.648],
    "tok_lulc@224": [0],
    "tok_dem@224": [670.665],
    "tok_ndvi@224": [0.327],
}

v1_pretraining_std = {
    "untok_sen2l2a@224": [2106.761, 2141.107, 2038.973, 2134.138, 2085.321, 1889.926, 1820.257, 1871.918, 1753.829, 1797.379, 1434.261, 1334.311],
    "untok_sen2l1c@224": [1624.683, 1675.806, 1557.708, 1833.702, 1823.738, 1733.977, 1732.131, 1679.732, 1727.26, 1024.687, 442.165, 1331.411, 1160.419],
    "untok_sen2rgb@224": [58.767, 47.663, 42.631],
    "untok_sen1grd@224": [5.195, 5.890],
    "untok_sen1rtc@224": [4.391, 4.459],
    "untok_dem@224": [951.272],
    "tok_sen2l2a@224": [2106.761, 2141.107, 2038.973, 2134.138, 2085.321, 1889.926, 1820.257, 1871.918, 1753.829, 1797.379, 1434.261, 1334.311],
    "tok_sen1grd@224": [5.195, 5.890],
    "tok_sen1rtc@224": [4.391, 4.459],
    "tok_lulc@224": [1],
    "tok_dem@224": [951.272],
    "tok_ndvi@224": [0.322],
}

v1_5_pretraining_mean = {**v1_pretraining_mean,
                         "naip@512": [107.20314516608458, 112.38461365995144, 93.28977241378502, 133.816892061648],
                         "untok_lulc@224": [0],
                         "untok_ndvi@224": [0.327],
                         "tok_sen2l1c@224": [2357.089, 2137.385, 2018.788, 2082.986, 2295.651, 2854.537, 3122.849,
                                             3040.56, 3306.481, 1473.847, 506.07, 2472.825, 1838.929],
                         "tok_sen2rgb@224": [87.271, 80.931, 66.667]
}
v1_5_pretraining_std = {**v1_pretraining_std,
                        "naip@512": [47.33802657655029, 39.109384320368385, 34.850792207554846, 47.32554578567598],
                        "untok_lulc@224": [1],
                        "untok_ndvi@224": [0.322],
                        "tok_sen2l1c@224": [1624.683, 1675.806, 1557.708, 1833.702, 1823.738, 1733.977, 1732.131,
                                            1679.732, 1727.26, 1024.687, 442.165, 1331.411, 1160.419],
                        "tok_sen2rgb@224": [58.767, 47.663, 42.631],
}


tokenizer_dict = {
    "v1": {
        "tok_sen2l2a@224": terramind_v1_tokenizer_s2l2a,
        "tok_sen1rtc@224": terramind_v1_tokenizer_s1rtc,
        "tok_sen1grd@224": terramind_v1_tokenizer_s1grd,
        "tok_dem@224": terramind_v1_tokenizer_dem,
        "tok_lulc@224": terramind_v1_tokenizer_lulc,
        "tok_ndvi@224": terramind_v1_tokenizer_ndvi,
        "coords": terramind_v1_coords_tokenizer,
    },
    "v01": {
        "tok_sen2l2a@224": terramind_v01_tokenizer_s2l2a,
        "tok_sen1grd@224": terramind_v01_tokenizer_s1grd,
        "tok_dem@224": terramind_v01_tokenizer_dem,
        "tok_lulc@224": terramind_v01_tokenizer_lulc,
        "coords": terramind_v1_coords_tokenizer,
        "caption": terramind_v01_caption_tokenizer,
    }
}


def select_modality_patch_embed_weights(model: TerraMindViT, bands: dict[str, list], pretrained_bands: dict[str, list]):
    """
    Update patch embeddings weights for each provided modality by selecting the pretrained weights for each band.
    Args:
         model (TerraMindViT): model
         bands (dict[str, list]): Bands with format {<modality>: [<band names>]}
         pretrained_bands (dict[str, list]): Pretrained bands of the model with format {<modality>: [<band names>]}
    """
    # Update modality names to match model layer names
    bands = {model.mod_name_mapping[k]: v for k, v in bands.items()}
    for mod, mod_bands in bands.items():
        if mod not in pretrained_bands:
            logger.info(f"Cannot load band weights for modality {mod}, not found in pretrained bands.")
            continue

        pixel_count = model.encoder_embeddings[mod].patch_size[0] * model.encoder_embeddings[mod].patch_size[1]

        pretrained_weight = model.encoder_embeddings[mod].proj.weight.clone()
        # Init new projection layer with updated number of channels
        model.encoder_embeddings[mod].proj = nn.Linear(
            pixel_count * len(mod_bands),
            model.encoder_embeddings[mod].dim_tokens,
            bias=False
        )
        temp_weight = model.encoder_embeddings[mod].proj.weight.clone()

        # Reshape to [dim, pixel, band]
        temp_weight = temp_weight.view(temp_weight.shape[0], pixel_count, -1)
        pretrained_weight = pretrained_weight.view(pretrained_weight.shape[0], pixel_count, -1)

        # Copy weights of bands
        for index, band in enumerate(mod_bands):
            if band in pretrained_bands[mod]:
                logging.info(f"Loaded weights for {band} in position {index} of patch embed")
                pretrained_index = pretrained_bands[mod].index(band)
                temp_weight[..., index] = pretrained_weight[..., pretrained_index]

        # Update model weights
        model.encoder_embeddings[mod].proj.weight = nn.Parameter(temp_weight.view(temp_weight.shape[0], -1))

    return model


def checkpoint_filter_fn(state_dict, model: TerraMindViT | TerraMind) -> dict:
    """Manually filter pre-trained weights for TerraMind to enable strict weight loading."""

    model_state_dict = model.state_dict()
    clean_dict = {}
    for k, v in state_dict.items():
        if k in model_state_dict:
            if v.shape == model_state_dict[k].shape:
                clean_dict[k] = v
            else:
                logger.warning(f"Shape for {k} ({list(v.shape)}) does not match model weights "
                               f"({list(model_state_dict[k].shape)}), skipping weights.")

    missing_params = set(model_state_dict.keys()) - set(clean_dict.keys())
    for k in missing_params:
        if not k.startswith('tokenizer'):
            logger.warning(f"Weights for {k} are missing in state dict, using random initialization.")
        clean_dict[k] = model_state_dict[k]

    state_dict = clean_dict

    return state_dict


def checkpoint_filter_fn_tim(state_dict, model: TerraMindTiM) -> dict:
    """Manually filter pre-trained weights for TerraMind ViT to enable strict weight loading."""

    model_state_dict = model.state_dict()
    clean_dict = {}
    for k, v in state_dict.items():
        if k in model_state_dict:
            if v.shape == model_state_dict[k].shape:
                clean_dict[k] = v
            else:
                logger.warning(f"Shape for {k} ({list(v.shape)}) does not match model weights "
                               f"({list(model_state_dict[k].shape)}), skipping weights.")
        if "sampler.model." + k in model_state_dict:
            # Copy weights for MAE model for TiM
            encdec_k = "sampler.model." + k
            if v.shape == model_state_dict[encdec_k].shape:
                clean_dict[encdec_k] = v
            else:
                raise ValueError(f"Shape for {k} ({list(v.shape)}) does not match MAE model weights "
                                 f"({list(model_state_dict[encdec_k].shape)}). Cannot run chain of thoughts without MAE.")

    missing_params = set(model_state_dict.keys()) - set(clean_dict.keys())
    for k in missing_params:
        if k.startswith("sampler.model."):
            raise ValueError(f"Weights for {k} are missing in state dict, cannot run chain of thoughts without MAE.")
        if not k.startswith('tokenizer'):
            logger.warning(f"Weights for {k} are missing in state dict, using random initialization.")
        clean_dict[k] = model_state_dict[k]

    state_dict = clean_dict

    return state_dict


def checkpoint_filter_fn_generate(state_dict, model: TerraMindGeneration) -> dict:
    """Manually filter pre-trained weights for TerraMind to enable strict weight loading."""

    model_state_dict = model.state_dict()
    clean_dict = {}
    for k, v in state_dict.items():
        encdec_k = "sampler.model." + k
        if encdec_k in model_state_dict:
            if v.shape == model_state_dict[encdec_k].shape:
                clean_dict[encdec_k] = v
            else:
                logger.warning(f"Shape for {k} ({list(v.shape)}) does not match model weights "
                               f"({list(model_state_dict[encdec_k].shape)}), skipping weights.")

    missing_params = set(model_state_dict.keys()) - set(clean_dict.keys())
    for k in missing_params:
        if not k.startswith("tokenizer"):
            # No warning because tokenizer weights are loaded separately
            logger.warning(f"Weights for {k} are missing in state dict, using random initialization.")
        clean_dict[k] = model_state_dict[k]

    state_dict = clean_dict

    return state_dict


def build_terrammind_vit(
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        bands: dict[str, list] | None = None,
        pretrained_bands: dict[str, list] | None = None,
        **kwargs):

    model = TerraMindViT(pretrained=pretrained, **kwargs)

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn(state_dict, model)
        model.load_state_dict(state_dict, strict=True)

    if bands is not None:
        model = select_modality_patch_embed_weights(model, bands, pretrained_bands)

    return model


def build_terrammind_encdec(
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        **kwargs):

    model = TerraMind(**kwargs)

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn(state_dict, model)
        model.load_state_dict(state_dict, strict=True)

    return model


def build_terrammind_tim(
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        bands: dict[str, list] | None = None,
        pretrained_bands: dict[str, list] | None = None,
        **kwargs):

    model = TerraMindTiM(pretrained=pretrained, **kwargs)

    if bands is not None:
        raise NotImplementedError(f"Bands cannot be adapted for TerraMind TiM models and is expected to be None. "
                                  f"Only exact matches with pre-trained modalities are supported, got {bands}.")

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        if any(isinstance(m, dict) for m in kwargs["modalities"]):
            raise NotImplementedError(f"TerraMind TiM models do not support new modalities. "
                                      f"Only pre-trained modalities are supported, got {kwargs['modalities']}")
            # TODO: Adapt model code to only use pre-trained modalities for TiM generation

        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn_tim(state_dict, model)
        model.load_state_dict(state_dict, strict=True)
    else:
        warnings.warn("TerraMind TiM model not pre-trained. Generation of TiM modalities will not work correctly.")

    return model


def build_terrammind_generate(
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        **kwargs):

    model = TerraMindGeneration(pretrained=pretrained, **kwargs)

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn_generate(state_dict, model)
        model.load_state_dict(state_dict, strict=True)

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_base(**kwargs):
    model = build_terrammind_vit(
        variant="terramind_v1_base",
        encoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model

      
@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_base_tim(**kwargs):
    model = build_terrammind_tim(
        variant="terramind_v1_base",
        encoder_depth=12,
        decoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v01_base(**kwargs):
    if kwargs.get('pretrained', False):
        if not os.getenv('HF_TOKEN', None):
            warnings.warn('TerraMind v0.1 models require a HF_TOKEN with access to model weights.')

    model = build_terrammind_vit(
        variant="terramind_v01_base",
        encoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands={"untok_sen2l2a@224": PRETRAINED_BANDS["untok_sen2l2a@224"]},
        tokenizer_dict=tokenizer_dict['v01'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v01_base_tim(**kwargs):
    model = build_terrammind_tim(
        variant="terramind_v01_base",
        encoder_depth=12,
        decoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands={"untok_sen2l2a@224": PRETRAINED_BANDS["untok_sen2l2a@224"]},
        tokenizer_dict=tokenizer_dict["v01"],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_large(**kwargs):
    model = build_terrammind_vit(
        variant="terramind_v1_large",
        encoder_depth=24,
        dim=1024,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict["v1"],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_large_tim(**kwargs):
    model = build_terrammind_tim(
        variant="terramind_v1_large",
        encoder_depth=24,
        decoder_depth=24,
        dim=1024,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict["v1"],
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_base_encdec(**kwargs):
    assert "encoder_embeddings" in kwargs and "decoder_embeddings" in kwargs and "modality_info" in kwargs, \
        ("TerraMind encdec models expect encoder_embeddings, decoder_embeddings, and modality_info. "
         "For generation, use the terramind_v1_base_generate model.")

    model = build_terrammind_encdec(
        variant="terramind_v1_base",
        encoder_depth=12,
        decoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_large_encdec(**kwargs):
    assert "encoder_embeddings" in kwargs and "decoder_embeddings" in kwargs and "modality_info" in kwargs, \
        ("TerraMind encdec models expect encoder_embeddings, decoder_embeddings, and modality_info. "
         "For generation, use the terramind_v1_large_generate model.")

    model = build_terrammind_encdec(
        variant="terramind_v1_large",
        encoder_depth=24,
        decoder_depth=24,
        dim=1024,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v01_base_generate(**kwargs):
    if kwargs.get('pretrained', False):
        if not os.getenv('HF_TOKEN', None):
            warnings.warn('TerraMind v0.1 models require a HF_TOKEN with access to model weights.')

    model = build_terrammind_generate(
        variant="terramind_v01_base",
        encoder_depth=12,
        decoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretraining_mean=v01_pretraining_mean,
        pretraining_std=v01_pretraining_std,
        tokenizer_dict=tokenizer_dict['v01'],
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_base_generate(**kwargs):
    model = build_terrammind_generate(
        variant="terramind_v1_base",
        encoder_depth=12,
        decoder_depth=12,
        dim=768,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretraining_mean=v1_pretraining_mean,
        pretraining_std=v1_pretraining_std,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_large_generate(**kwargs):
    model = build_terrammind_generate(
        variant="terramind_v1_large",
        encoder_depth=24,
        decoder_depth=24,
        dim=1024,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=False,
        proj_bias=False,
        mlp_bias=False,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.SiLU,
        gated_mlp=True,
        pretraining_mean=v1_pretraining_mean,
        pretraining_std=v1_pretraining_std,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tiny(**kwargs):
    model = build_terrammind_vit(
        variant="terramind_v1_tiny",
        encoder_depth=12,
        dim=192,
        num_heads=3,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tiny_tim(**kwargs):
    model = build_terrammind_tim(
        variant="terramind_v1_tiny",
        encoder_depth=12,
        decoder_depth=4,
        dim=192,
        num_heads=3,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tiny_encdec(**kwargs):
    model = build_terrammind_encdec(
        variant="terramind_v1_tiny",
        encoder_depth=12,
        decoder_depth=4,
        dim=192,
        num_heads=3,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tiny_generate(**kwargs):
    model = build_terrammind_generate(
        variant="terramind_v1_tiny",
        encoder_depth=12,
        decoder_depth=4,
        dim=192,
        num_heads=3,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretraining_mean=v1_pretraining_mean,
        pretraining_std=v1_pretraining_std,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_small(**kwargs):
    model = build_terrammind_vit(
        variant="terramind_v1_small",
        encoder_depth=12,
        dim=384,
        num_heads=6,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_small_tim(**kwargs):
    model = build_terrammind_tim(
        variant="terramind_v1_small",
        encoder_depth=12,
        decoder_depth=6,
        dim=384,
        num_heads=6,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_small_encdec(**kwargs):
    model = build_terrammind_encdec(
        variant="terramind_v1_small",
        encoder_depth=12,
        decoder_depth=6,
        dim=384,
        num_heads=6,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretrained_bands=PRETRAINED_BANDS,
        **kwargs
    )
    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_small_generate(**kwargs):
    model = build_terrammind_generate(
        variant="terramind_v1_small",
        encoder_depth=12,
        decoder_depth=6,
        dim=384,
        num_heads=6,
        mlp_ratio=4,
        qkv_bias=True,
        proj_bias=True,
        mlp_bias=True,
        num_register_tokens=0,
        norm_layer=partial(LayerNorm, eps=1e-6, bias=False),
        act_layer=nn.GELU,
        gated_mlp=False,
        pretraining_mean=v1_pretraining_mean,
        pretraining_std=v1_pretraining_std,
        tokenizer_dict=tokenizer_dict['v1'],
        **kwargs
    )
    return model



"""
Code from https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/simple_vit.py

"""

import torch


def posemb_sincos_2d(h, w, dim, temperature: int = 10000, dtype=torch.float32):
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
    assert (dim % 4) == 0, "feature dimension must be multiple of 4 for sincos emb"
    omega = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature**omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]
    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)
    return pe.type(dtype)


def posemb_sincos_2d_with_gsd(
    h, w, dim, gsd=1.0, temperature: int = 10000, dtype=torch.float32
):
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
    assert (dim % 4) == 0, "feature dimension must be multiple of 4 for sincos emb"

    gsd = gsd.to(x.device)
    omega = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature ** (2 * omega / dim)) * (gsd / 1.0)  # Adjusted for g

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]
    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)
    return pe.type(dtype)


def posemb_sincos_1d(waves, dim, temperature: int = 10000, dtype=torch.float32):
    assert (
        dim % 2 == 0
    ), "Feature dimension must be a multiple of 2 for sincos embedding"
    waves = torch.arange(waves) if isinstance(waves, int) else waves

    omega = torch.arange(dim // 2, device=waves.device) / (dim // 2 - 1)
    omega = 1.0 / (temperature**omega)

    scaled_waves = waves[:, None] * omega[None, :]
    pe = torch.cat((scaled_waves.sin(), scaled_waves.cos()), dim=1)

    return pe.type(dtype)
